# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

import torch
from torch import Tensor

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import Backend, get_backend, get_ir_config


# Here using mmcv.ops.roi_align.__self__ to find
# mmcv.ops.roi_align.RoIAlignFunction, because RoIAlignFunction is not
# visible in mmcv.
@SYMBOLIC_REWRITER.register_symbolic("mmcv.ops.roi_align.__self__", backend="default")
def roi_align_default(
    g,
    input: Tensor,
    rois: Tensor,
    output_size: List[int],
    spatial_scale: float,
    sampling_ratio: int,
    pool_mode: str,
    aligned: bool,
):
    """Rewrite symbolic function for default backend.

    Replace onnx::RoiAlign with mmcv::MMCVRoiAlign for PPLNN. For ONNXRuntime,
    align operation get done outside the inference engine for opset versions
    lower than 16. By default,  onnx::RoiAlign get replaced to
    mmdeploy::MMCVRoiAlign.

    Args:
        ctx (ContextCaller): The context with additional information.
        g (Graph): The traced onnx graph.
        input (Tensor): Input tensor, 4-D feature map of shape (N, C, H, W).
        rois (Tensor): Bx5 boxes. First column is the index into N. The other
            4 columns are xyxy.
        output_size(List[int]): Output size of height and width.
        spatial_scale (float):
        sampling_ratio (int): Number of inputs samples to take for each
            output sample. 0 to take samples densely for current models.
        pool_mode (str): Pooling mode in each bin, could be 'avg' or 'max'.
        aligned (bool): With `aligned=True`, we first appropriately scale
            the ROI and then shift it by -0.5 prior to calling roi_align.
            This produces the correct neighbors;

    Returns:
        MMCVRoiAlign op for onnx.
    """
    ctx = SYMBOLIC_REWRITER.get_context()
    backend = get_backend(ctx.cfg)
    if backend == Backend.PPLNN or backend == Backend.TENSORRT:
        domain = "mmcv"
        return g.op(
            f"{domain}::MMCVRoiAlign",
            input,
            rois,
            output_height_i=output_size[0],
            output_width_i=output_size[1],
            spatial_scale_f=spatial_scale,
            sampling_ratio_i=sampling_ratio,
            mode_s=pool_mode,
            aligned_i=aligned,
        )
    else:
        from torch.onnx.symbolic_opset9 import _cast_Long
        from torch.onnx.symbolic_opset11 import add, select

        batch_indices = _cast_Long(
            g,
            g.op(
                "Squeeze",
                select(
                    g,
                    rois,
                    1,
                    g.op("Constant", value_t=torch.tensor([0], dtype=torch.long)),
                ),
                axes_i=[1],
            ),
            False,
        )
        rois = select(
            g,
            rois,
            1,
            g.op("Constant", value_t=torch.tensor([1, 2, 3, 4], dtype=torch.long)),
        )
        ir_cfg = get_ir_config(ctx.cfg)
        opset_version = ir_cfg.get("opset_version", 11)
        if opset_version < 16:
            # preprocess rois to make compatible with opset 16-
            # as for opset 16+, `aligned` get implemented inside onnxruntime.
            if aligned is True:
                rois = add(
                    g,
                    rois,
                    g.op(
                        "Constant",
                        value_t=torch.tensor([-0.5 / spatial_scale], dtype=torch.float),
                    ),
                )
            return g.op(
                "RoiAlign",
                input,
                rois,
                batch_indices,
                output_height_i=output_size[0],
                output_width_i=output_size[1],
                spatial_scale_f=spatial_scale,
                sampling_ratio_i=sampling_ratio,
                mode_s=pool_mode,
            )
        else:
            return g.op(
                "RoiAlign",
                input,
                rois,
                batch_indices,
                output_height_i=output_size[0],
                output_width_i=output_size[1],
                spatial_scale_f=spatial_scale,
                sampling_ratio_i=sampling_ratio,
                mode_s=pool_mode,
                aligned_i=aligned,
            )



"""Main class to perform inference using RTMDet and RTMPose (ONNX)"""

import cv2
import csv
from pathlib import Path
import numpy as np
import supervision as sv

from typing import List, Union, Generator, Literal, Optional

from juxtapose.data import load_inference_source
from juxtapose.detectors import get_detector, RTMDet, GroundingDino, YOLOv8
from juxtapose.rtmpose import RTMPose

from juxtapose.utils.core import Detections
from juxtapose.utils.plotting import Annotator
from juxtapose.utils.roi import select_roi
from juxtapose.utils.checks import check_imshow
from juxtapose.utils.torch_utils import smart_inference_mode
from juxtapose.utils.ops import xyxy2xyxyxyxy
from juxtapose.utils import (
    LOGGER,
    MACOS,
    WINDOWS,
    colorstr,
    ops,
    get_time,
)
from juxtapose.trackers import Tracker, TRACKER_MAP

from pydantic import BaseModel


class Result(BaseModel):
    im: np.ndarray  # shape -> (h, w, c)
    kpts: List  # shape -> (number of humans, 17, 2)
    bboxes: List  # shape -> (number of humans, 4)
    speed: dict  # {'bboxes': ... ms, 'kpts': ... ms} -> used to record the milliseconds of the inference time

    save_dirs: str  # save directory
    name: str  # file name

    class Config:
        arbitrary_types_allowed = True


class RTM:
    """"""

    def __init__(
        self,
        device="cpu",
        annotator=Annotator(),
    ) -> None:
        self.dataset = None
        self.vid_path, self.vid_writer = None, None

    def setup_source(self, source, imgsz=640, vid_stride=1) -> None:
        self.dataset = load_inference_source(
            source=source, imgsz=imgsz, vid_stride=vid_stride
        )
        self.source_type = self.dataset.source_type
        self.vid_path, self.vid_writer = [None] * self.dataset.bs, [
            None
        ] * self.dataset.bs

    def save_preds(self, im0, vid_cap, idx, save_path) -> None:
        """Save video predictions as mp4 at specified path."""
        # Save imgs
        if self.dataset.mode == "image":
            cv2.imwrite(save_path, im0)
        else:  # 'video' or 'stream'
            if self.vid_path[idx] != save_path:  # new video
                self.vid_path[idx] = save_path
                if isinstance(self.vid_writer[idx], cv2.VideoWriter):
                    self.vid_writer[idx].release()  # release previous video writer
                if vid_cap:  # video
                    fps = int(
                        vid_cap.get(cv2.CAP_PROP_FPS)
                    )  # integer required, floats produce error in MP4 codec
                    w = int(vid_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                    h = int(vid_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                else:  # stream
                    fps, w, h = 30, im0.shape[1], im0.shape[0]
                suffix = ".mp4" if MACOS else ".avi" if WINDOWS else ".avi"
                fourcc = "avc1" if MACOS else "WMV2" if WINDOWS else "MJPG"
                save_path = str(Path(save_path).with_suffix(suffix))
                self.vid_writer[idx] = cv2.VideoWriter(
                    save_path, cv2.VideoWriter_fourcc(*fourcc), fps, (w, h)
                )
            self.vid_writer[idx].write(im0)

    def save_csv(self, path, data):
        with open(path, "a", newline="", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerow(data)

    def get_labels(self, detections: Detections):
        mp = {0: "person"}
        return np.array(
            [
                f"{mp[label]} {id} {score:.2f}"
                for score, label, id in zip(
                    detections.confidence, detections.labels, detections.track_id
                )
            ]
        )

    @smart_inference_mode
    def stream_inference(
        self,
        source,
        # interest point
        timer: List = [],
        poi: Literal["point", "box", "text", ""] = "",
        roi: Literal["rect", ""] = "",
        # panels
        show=True,
        plot=True,
        plot_bboxes=True,
        save=False,
        save_dirs="",
        verbose=True,
    ) -> Result:
        if show:
            check_imshow(warn=True)

        if save:
            save_dirs = Path(get_time() if not save_dirs else save_dirs)
            save_dirs.mkdir(exist_ok=True)

        profilers = (ops.Profile(), ops.Profile(), ops.Profile())  # count the time
        self.setup_source(source)

        if self.tracker_type:
            self.setup_tracker()

        current_source, index = None, 0

        for _, batch in enumerate(self.dataset):
            path, im0s, vid_cap, s = batch
            is_image = s.startswith("image")

            p, im, im0 = Path(path[0]), im0s[0], im0s[0].copy()

            index += 1

            # reset tracker when source changed
            if current_source is None:
                current_source = p
                if roi:
                    zones = self.setup_roi(im)
            elif current_source != p:
                if roi:
                    zones = self.setup_roi(im)
                if self.tracker_type:
                    self.setup_tracker()
                current_source = p
                index = 1

            # multi object detection (detect only person)
            with profilers[0]:
                detections: Detections = self.det(im)

            if plot:
                labels = self.get_labels(detections)

                if plot_bboxes:
                    self.box_annotator.annotate(im, detections, labels=labels)
                # self.annotator.draw_kpts(im, kpts)
                # self.annotator.draw_skeletons(im, kpts)

            result = Result(
                im=im,
                bboxes=detections.xyxy,  # detections.xyxy,
                speed={
                    "bboxes": profilers[0].dt * 1e3 / 1,
                    "track": profilers[1].dt * 1e3 / 1,
                    "kpts": profilers[2].dt * 1e3 / 1,
                },
                save_dirs=str(save_dirs) if save else "",
                name=p.name,
            )

            yield result

            if show:
                if is_image:
                    cv2.imshow(result.name, result.im)
                    key = cv2.waitKey(0) & 0xFF
                    if key == ord("q"):
                        cv2.destroyWindow(result.name)
                        continue
                else:
                    cv2.imshow(result.name, result.im)
                    key = cv2.waitKey(1) & 0xFF
                    if key == ord("q"):
                        break
                    elif key == 32:
                        zones = self.setup_roi(im0)

            if verbose:
                LOGGER.info(
                    f"{s}Detected {len(detections.xyxy)} person(s) ({colorstr('green', f'{profilers[0].dt * 1E3:.1f}ms')} | {colorstr('green', f'{profilers[1].dt * 1E3:.1f}ms')} | {colorstr('green', f'{profilers[2].dt * 1E3:.1f}ms')})"
                )

            if save:
                self.save_preds(im, vid_cap, 0, str(save_dirs / p.name))
                self.save_csv(
                    str(save_dirs / p.with_suffix(".csv").name),
                    [
                        str(index),
                        # str([{i: kpt} for i, kpt in zip(detections.track_id, kpts)]),
                    ],
                )

        if isinstance(self.vid_writer[-1], cv2.VideoWriter):
            self.vid_writer[-1].release()

        cv2.destroyAllWindows()

    def __call__(
        self,
        source,
        stream=False,
        # interest point
        timer: List = [],
        poi: Literal["point", "box", "text", ""] = "",
        roi: Literal["rect", ""] = "",
        # panels
        show=True,
        plot=True,
        plot_bboxes=True,
        save=False,
        save_dirs="",
        verbose=True,
    ) -> Union[List[Result], Generator[Result, None, None]]:
        if stream:
            return self.stream_inference(
                source,
                timer=timer,
                poi=poi,
                roi=roi,
                show=show,
                plot=plot,
                plot_bboxes=plot_bboxes,
                save=save,
                save_dirs=save_dirs,
                verbose=verbose,
            )
        else:
            return list(
                self.stream_inference(
                    source,
                    timer=timer,
                    poi=poi,
                    roi=roi,
                    show=show,
                    plot=plot,
                    plot_bboxes=plot_bboxes,
                    save=save,
                    save_dirs=save_dirs,
                    verbose=verbose,
                )
            )



# Copyright (c) OpenMMLab. All rights reserved.
from . import backbones  # noqa: F401, F403
from . import dense_heads  # noqa: F401,F403
from . import detectors  # noqa: F401,F403
from . import layers  # noqa: F401,F403
from . import necks  # noqa: F401,F403
from . import roi_heads  # noqa: F401,F403
from . import task_modules  # noqa: F401,F403
from . import transformer  # noqa: F401,F403
from . import utils  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Tuple, Union

import torch
from torch.autograd import Function

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend, get_root_logger


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.interpolate", backend="ncnn"
)
def interpolate__ncnn(
    input: torch.Tensor,
    size: Optional[
        Union[int, Tuple[int], Tuple[int, int], Tuple[int, int, int]]
    ] = None,
    scale_factor: Optional[Union[float, Tuple[float]]] = None,
    mode: str = "nearest",
    align_corners: Optional[bool] = None,
    recompute_scale_factor: Optional[bool] = None,
):
    """Rewrite `interpolate` for ncnn backend.

    ncnn require `size` should be constant in ONNX Node. We use `scale_factor`
    instead of `size` to avoid dynamic size.
    """
    ctx = FUNCTION_REWRITER.get_context()

    input_size = input.shape
    if scale_factor is None:
        scale_factor = [s_out / s_in for s_out, s_in in zip(size, input_size[2:])]

    return ctx.origin_func(
        input,
        None,
        scale_factor,
        mode=mode,
        align_corners=align_corners,
        recompute_scale_factor=recompute_scale_factor,
    )


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.interpolate", backend="rknn"
)
def interpolate__rknn(
    input: torch.Tensor,
    size: Optional[
        Union[int, Tuple[int], Tuple[int, int], Tuple[int, int, int]]
    ] = None,
    scale_factor: Optional[Union[float, Tuple[float]]] = None,
    mode: str = "nearest",
    align_corners: Optional[bool] = None,
    recompute_scale_factor: Optional[bool] = None,
):
    """Rewrite `interpolate` for rknn backend.

    rknn require `size` should be constant in ONNX Node. We use `scale_factor`
    instead of `size` to avoid dynamic size.
    """
    ctx = FUNCTION_REWRITER.get_context()
    input_size = input.shape
    if scale_factor is None:
        scale_factor = [(s_out / s_in) for s_out, s_in in zip(size, input_size[2:])]
        if isinstance(scale_factor[0], torch.Tensor):
            scale_factor = [i.item() for i in scale_factor]

    return ctx.origin_func(
        input,
        None,
        scale_factor,
        mode=mode,
        align_corners=align_corners,
        recompute_scale_factor=recompute_scale_factor,
    )


@FUNCTION_REWRITER.register_rewriter(
    "torch.nn.functional.interpolate", is_pytorch=True, backend=Backend.TENSORRT.value
)
def interpolate__tensorrt(
    input: torch.Tensor,
    size: Optional[
        Union[int, Tuple[int], Tuple[int, int], Tuple[int, int, int]]
    ] = None,
    scale_factor: Optional[Union[float, Tuple[float]]] = None,
    mode: str = "nearest",
    align_corners: Optional[bool] = None,
    recompute_scale_factor: Optional[bool] = None,
):
    """Register default symbolic function for `interpolate`."""
    ctx = FUNCTION_REWRITER.get_context()

    class BicubicInterpolate(Function):
        def __init__(self) -> None:
            super().__init__()

        @staticmethod
        def symbolic(g, input, scale_factor, align_corners):
            """Symbolic function for creating onnx op."""
            return g.op(
                "mmdeploy::TRTBicubicInterpolate",
                input,
                scale_factor_f=scale_factor,
                align_corners_i=align_corners,
            )

        @staticmethod
        def forward(g, input, scale_factor, align_corners):
            """Run forward."""
            return ctx.origin_func(
                input,
                scale_factor=scale_factor,
                mode="bicubic",
                align_corners=align_corners,
            )

    if "bicubic" == mode:
        input_size = input.shape
        if isinstance(scale_factor, float):
            scale_factor = [scale_factor, scale_factor]
        if scale_factor is None:
            logger = get_root_logger()
            logger.warning(
                "ResizeLayer in TensorRT allow dynamic input shape with shape "
                "tensor. Which is not available for custom ops. Computed scale"
                "_factor might be the right way to get final shape."
            )
            scale_factor = [
                float(s_out / s_in) for s_out, s_in in zip(size, input_size[2:])
            ]
        return BicubicInterpolate.apply(input, scale_factor, align_corners)
    else:
        return ctx.origin_func(
            input,
            size=size,
            scale_factor=scale_factor,
            mode=mode,
            align_corners=align_corners,
            recompute_scale_factor=recompute_scale_factor,
        )



# Copyright (c) OpenMMLab. All rights reserved.
from . import deform_conv  # noqa: F401,F403
from . import modulated_deform_conv  # noqa: F401,F403
from . import multi_scale_deform_attn  # noqa: F401,F403
from . import point_sample  # noqa: F401,F403
from . import roi_align  # noqa: F401,F403
from . import roi_align_rotated  # noqa: F401,F403
from . import transformer  # noqa: F401,F403
from .nms import ONNXNMSop, TRTBatchedNMSop, multiclass_nms
from .nms_rotated import (ONNXNMSRotatedOp, TRTBatchedRotatedNMSop,
                          multiclass_nms_rotated)

__all__ = [
    'ONNXNMSop', 'TRTBatchedNMSop', 'TRTBatchedRotatedNMSop',
    'ONNXNMSRotatedOp', 'multiclass_nms', 'multiclass_nms_rotated'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


def update_squeeze_unsqueeze_opset13_pass(graph, params_dict, torch_out):
    """Update Squeeze/Unsqueeze axes for opset13."""
    for node in graph.nodes():
        if node.kind() in ["onnx::Squeeze", "onnx::Unsqueeze"] and node.hasAttribute(
            "axes"
        ):
            axes = node["axes"]
            axes_node = graph.create("onnx::Constant")
            axes_node.t_("value", torch.LongTensor(axes))
            node.removeAttribute("axes")
            node.addInput(axes_node.output())
            axes_node.insertBefore(node)
    return graph, params_dict, torch_out


@FUNCTION_REWRITER.register_rewriter("torch.onnx.utils._model_to_graph")
def model_to_graph__custom_optimizer(*args, **kwargs):
    """Rewriter of _model_to_graph, add custom passes."""
    ctx = FUNCTION_REWRITER.get_context()
    graph, params_dict, torch_out = ctx.origin_func(*args, **kwargs)
    if hasattr(ctx, "opset"):
        opset_version = ctx.opset
    else:
        from juxtapose.mmdeploy.utils import get_ir_config

        opset_version = get_ir_config(ctx.cfg).get("opset_version", 11)
    if opset_version >= 13:
        graph, params_dict, torch_out = update_squeeze_unsqueeze_opset13_pass(
            graph, params_dict, torch_out
        )
    custom_passes = getattr(ctx, "onnx_custom_passes", None)

    if custom_passes is not None:
        assert isinstance(
            custom_passes, Callable
        ), f"Expect a callable onnx_custom_passes, get {type(custom_passes)}."
        graph, params_dict, torch_out = custom_passes(
            ctx, graph, params_dict, torch_out
        )

    return graph, params_dict, torch_out


@FUNCTION_REWRITER.register_rewriter(
    "torch._C._jit_pass_onnx_deduplicate_initializers", backend="tensorrt"
)
def jit_pass_onnx_deduplicate_initializers__disable(graph, param_dict, arg2):
    """This pass will disable TensorRT topk export.

    disable for TensorRT.
    """
    return param_dict


@FUNCTION_REWRITER.register_rewriter(
    "torch._C._jit_pass_onnx_autograd_function_process"
)
def jit_pass_onnx_autograd_function_process__disable(graph):
    """Disable process autograph function."""
    return



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.utils import get_root_logger


def optimize_onnx(ctx, graph, params_dict, torch_out):
    """The optimize callback of the onnx model."""
    logger = get_root_logger()
    logger.info("Execute onnx optimize passes.")
    try:
        from juxtapose.mmdeploy.backend.torchscript import ts_optimizer

        ts_optimizer.onnx._jit_pass_merge_shape_concate(graph)
        ts_optimizer.onnx._jit_pass_onnx_peephole(graph)
        ts_optimizer.onnx._jit_pass_flatten_cls_head(graph)
        ts_optimizer.onnx._jit_pass_fuse_select_assign(graph, params_dict)
        ts_optimizer.onnx._jit_pass_common_subgraph_elimination(graph, params_dict)
    except ImportError:
        logger.warning(
            "Can not optimize model, please build torchscipt extension.\n"
            "More details: "
            "https://github.com/open-mmlab/mmdeploy/tree/main/docs/en/experimental/onnx_optimizer.md"  # noqa
        )
    return graph, params_dict, torch_out



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.any")
@FUNCTION_REWRITER.register_rewriter(func_name="torch.any")
def any__default(
    input: torch.Tensor, dim: Optional[str] = None, keepdim: bool = False, **kwargs
) -> torch.Tensor:
    """Rewrite `any` for ONNX."""
    if dim is None and keepdim is False:
        return (input != 0).sum() > 0

    return (input != 0).sum(dim, keepdim=keepdim) > 0



# ------------------------------------------------------------------------
# Grounding DINO
# url: https://github.com/IDEA-Research/GroundingDINO
# Copyright (c) 2023 IDEA. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------

import torch
import torch.nn.functional as F
import torch.utils.checkpoint as checkpoint
from torch import Tensor, nn
from torchvision.ops.boxes import nms
from transformers import BertConfig, BertModel, BertPreTrainedModel
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions


class BertModelWarper(nn.Module):
    def __init__(self, bert_model):
        super().__init__()
        # self.bert = bert_modelc

        self.config = bert_model.config
        self.embeddings = bert_model.embeddings
        self.encoder = bert_model.encoder
        self.pooler = bert_model.pooler

        self.get_extended_attention_mask = bert_model.get_extended_attention_mask
        self.invert_attention_mask = bert_model.invert_attention_mask
        self.get_head_mask = bert_model.get_head_mask

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        r"""
        encoder_hidden_states  (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, hidden_size)`, `optional`):
            Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention if
            the model is configured as a decoder.
        encoder_attention_mask (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Mask to avoid performing attention on the padding token indices of the encoder input. This mask is used in
            the cross-attention if the model is configured as a decoder. Mask values selected in ``[0, 1]``:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        past_key_values (:obj:`tuple(tuple(torch.FloatTensor))` of length :obj:`config.n_layers` with each tuple having 4 tensors of shape :obj:`(batch_size, num_heads, sequence_length - 1, embed_size_per_head)`):
            Contains precomputed key and value hidden states of the attention blocks. Can be used to speed up decoding.

            If :obj:`past_key_values` are used, the user can optionally input only the last :obj:`decoder_input_ids`
            (those that don't have their past key value states given to this model) of shape :obj:`(batch_size, 1)`
            instead of all :obj:`decoder_input_ids` of shape :obj:`(batch_size, sequence_length)`.
        use_cache (:obj:`bool`, `optional`):
            If set to :obj:`True`, :obj:`past_key_values` key value states are returned and can be used to speed up
            decoding (see :obj:`past_key_values`).
        """
        output_attentions = (
            output_attentions if output_attentions is not None else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if self.config.is_decoder:
            use_cache = use_cache if use_cache is not None else self.config.use_cache
        else:
            use_cache = False

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
            batch_size, seq_length = input_shape
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
            batch_size, seq_length = input_shape
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        # past_key_values_length
        past_key_values_length = (
            past_key_values[0][0].shape[2] if past_key_values is not None else 0
        )

        if attention_mask is None:
            attention_mask = torch.ones(
                ((batch_size, seq_length + past_key_values_length)), device=device
            )
        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(
            attention_mask, input_shape, device
        )

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if self.config.is_decoder and encoder_hidden_states is not None:
            encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
            if encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
            encoder_extended_attention_mask = self.invert_attention_mask(encoder_attention_mask)
        else:
            encoder_extended_attention_mask = None
        # if os.environ.get('IPDB_SHILONG_DEBUG', None) == 'INFO':
        #     import ipdb; ipdb.set_trace()

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
        )

        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


class TextEncoderShell(nn.Module):
    def __init__(self, text_encoder):
        super().__init__()
        self.text_encoder = text_encoder
        self.config = self.text_encoder.config

    def forward(self, **kw):
        # feed into text encoder
        return self.text_encoder(**kw)


def generate_masks_with_special_tokens(tokenized, special_tokens_list, tokenizer):
    """Generate attention mask between each pair of special tokens
    Args:
        input_ids (torch.Tensor): input ids. Shape: [bs, num_token]
        special_tokens_mask (list): special tokens mask.
    Returns:
        torch.Tensor: attention mask between each special tokens.
    """
    input_ids = tokenized["input_ids"]
    bs, num_token = input_ids.shape
    # special_tokens_mask: bs, num_token. 1 for special tokens. 0 for normal tokens
    special_tokens_mask = torch.zeros((bs, num_token), device=input_ids.device).bool()
    for special_token in special_tokens_list:
        special_tokens_mask |= input_ids == special_token

    # idxs: each row is a list of indices of special tokens
    idxs = torch.nonzero(special_tokens_mask)

    # generate attention mask and positional ids
    attention_mask = (
        torch.eye(num_token, device=input_ids.device).bool().unsqueeze(0).repeat(bs, 1, 1)
    )
    position_ids = torch.zeros((bs, num_token), device=input_ids.device)
    previous_col = 0
    for i in range(idxs.shape[0]):
        row, col = idxs[i]
        if (col == 0) or (col == num_token - 1):
            attention_mask[row, col, col] = True
            position_ids[row, col] = 0
        else:
            attention_mask[row, previous_col + 1 : col + 1, previous_col + 1 : col + 1] = True
            position_ids[row, previous_col + 1 : col + 1] = torch.arange(
                0, col - previous_col, device=input_ids.device
            )

        previous_col = col

    # # padding mask
    # padding_mask = tokenized['attention_mask']
    # attention_mask = attention_mask & padding_mask.unsqueeze(1).bool() & padding_mask.unsqueeze(2).bool()

    return attention_mask, position_ids.to(torch.long)


def generate_masks_with_special_tokens_and_transfer_map(tokenized, special_tokens_list, tokenizer):
    """Generate attention mask between each pair of special tokens
    Args:
        input_ids (torch.Tensor): input ids. Shape: [bs, num_token]
        special_tokens_mask (list): special tokens mask.
    Returns:
        torch.Tensor: attention mask between each special tokens.
    """
    input_ids = tokenized["input_ids"]
    bs, num_token = input_ids.shape
    # special_tokens_mask: bs, num_token. 1 for special tokens. 0 for normal tokens
    special_tokens_mask = torch.zeros((bs, num_token), device=input_ids.device).bool()
    for special_token in special_tokens_list:
        special_tokens_mask |= input_ids == special_token

    # idxs: each row is a list of indices of special tokens
    idxs = torch.nonzero(special_tokens_mask)

    # generate attention mask and positional ids
    attention_mask = (
        torch.eye(num_token, device=input_ids.device).bool().unsqueeze(0).repeat(bs, 1, 1)
    )
    position_ids = torch.zeros((bs, num_token), device=input_ids.device)
    cate_to_token_mask_list = [[] for _ in range(bs)]
    previous_col = 0
    for i in range(idxs.shape[0]):
        row, col = idxs[i]
        if (col == 0) or (col == num_token - 1):
            attention_mask[row, col, col] = True
            position_ids[row, col] = 0
        else:
            attention_mask[row, previous_col + 1 : col + 1, previous_col + 1 : col + 1] = True
            position_ids[row, previous_col + 1 : col + 1] = torch.arange(
                0, col - previous_col, device=input_ids.device
            )
            c2t_maski = torch.zeros((num_token), device=input_ids.device).bool()
            c2t_maski[previous_col + 1 : col] = True
            cate_to_token_mask_list[row].append(c2t_maski)
        previous_col = col

    cate_to_token_mask_list = [
        torch.stack(cate_to_token_mask_listi, dim=0)
        for cate_to_token_mask_listi in cate_to_token_mask_list
    ]

    # # padding mask
    # padding_mask = tokenized['attention_mask']
    # attention_mask = attention_mask & padding_mask.unsqueeze(1).bool() & padding_mask.unsqueeze(2).bool()

    return attention_mask, position_ids.to(torch.long), cate_to_token_mask_list



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import get_root_logger


@FUNCTION_REWRITER.register_rewriter(func_name="torch.topk", backend="default")
@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.topk", backend="default")
def topk__dynamic(
    input: torch.Tensor,
    k: int,
    dim: Optional[int] = None,
    largest: bool = True,
    sorted: bool = True,
):
    """Rewrite `topk` for default backend.

    Cast k to tensor and make sure k is smaller than input.shape[dim].
    """
    ctx = FUNCTION_REWRITER.get_context()

    if dim is None:
        dim = int(input.ndim - 1)
    size = input.shape[dim]
    if not isinstance(k, torch.Tensor):
        k = torch.tensor(k, device=input.device, dtype=torch.long)
    # Always keep topk op for dynamic input
    if isinstance(size, torch.Tensor):
        # size would be treated as cpu tensor, trick to avoid that.
        size = k.new_zeros(()) + size
    k = torch.where(k < size, k, size)
    return ctx.origin_func(input, k, dim=dim, largest=largest, sorted=sorted)


@FUNCTION_REWRITER.register_rewriter(func_name="torch.topk", backend="tensorrt")
@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.topk", backend="tensorrt")
def topk__tensorrt(
    input: torch.Tensor,
    k: int,
    dim: Optional[int] = None,
    largest: bool = True,
    sorted: bool = True,
):
    """Rewrite `topk` for TensorRT backend.

    TensorRT does not support topk with dynamic k. This function cast k to
    constant integer.
    """
    ctx = FUNCTION_REWRITER.get_context()
    # https://docs.nvidia.com/deeplearning/tensorrt/developer-guide/index.html#topKsetup
    from juxtapose.mmdeploy.utils.constants import TENSORRT_MAX_TOPK

    if dim is None:
        dim = int(input.ndim - 1)
    size = input.shape[dim]
    if k > size:
        k = size
    if not isinstance(k, int):
        k = int(k)
    if k > TENSORRT_MAX_TOPK:
        logger = get_root_logger()
        logger.warning(
            f"Maximum K of TopK in TensorRT is {TENSORRT_MAX_TOPK}, "
            f"but given {k}. Note that k will be set "
            f"to {TENSORRT_MAX_TOPK}."
        )
        k = TENSORRT_MAX_TOPK

    return ctx.origin_func(input, k, dim=dim, largest=largest, sorted=sorted)


@FUNCTION_REWRITER.register_rewriter(func_name="torch.topk", backend="coreml")
@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.topk", backend="coreml")
def topk__coreml(
    input: torch.Tensor,
    k: int,
    dim: Optional[int] = None,
    largest: bool = True,
    sorted: bool = True,
):
    """Rewrite `topk` for coreml.

    Cast k to tensor and make sure k is smaller than input.shape[dim].
    """
    ctx = FUNCTION_REWRITER.get_context()

    if dim is None:
        dim = int(input.ndim - 1)
    size = input.shape[dim]
    if not isinstance(k, torch.Tensor):
        k = torch.tensor(k, device=input.device, dtype=torch.long)
    # Always keep topk op for dynamic input
    k = torch.where(k < size, k, size)
    return ctx.origin_func(input, k, dim=dim, largest=largest, sorted=sorted)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Iterable, Optional, Union

import onnx
import onnx.helper
import onnx.utils

from juxtapose.mmdeploy.apis.core import PIPELINE_MANAGER
from juxtapose.mmdeploy.core.optimizers import (
    attribute_to_dict,
    create_extractor,
    get_new_name,
    parse_extractor_io_string,
    remove_identity,
    remove_imports,
    rename_value,
)
from juxtapose.mmdeploy.utils import get_root_logger


@PIPELINE_MANAGER.register_pipeline()
def extract_partition(
    model: Union[str, onnx.ModelProto],
    start_marker: Union[str, Iterable[str]],
    end_marker: Union[str, Iterable[str]],
    start_name_map: Optional[Dict[str, str]] = None,
    end_name_map: Optional[Dict[str, str]] = None,
    dynamic_axes: Optional[Dict[str, Dict[int, str]]] = None,
    save_file: Optional[str] = None,
) -> onnx.ModelProto:
    """Extract partition-model from an ONNX model.

    The partition-model is defined by the names of the input and output tensors
    exactly.

    Examples:
        >>> from juxtapose.mmdeploy.apis import extract_model
        >>> model = 'work_dir/fastrcnn.onnx'
        >>> start_marker = 'detector:input'
        >>> end_marker = ['extract_feat:output', 'multiclass_nms[0]:input']
        >>> dynamic_axes = {
            'input': {
                0: 'batch',
                2: 'height',
                3: 'width'
            },
            'scores': {
                0: 'batch',
                1: 'num_boxes',
            },
            'boxes': {
                0: 'batch',
                1: 'num_boxes',
            }
        }
        >>> save_file = 'partition_model.onnx'
        >>> extract_partition(model, start_marker, end_marker, \
                dynamic_axes=dynamic_axes, \
                save_file=save_file)

    Args:
        model (str | onnx.ModelProto): Input ONNX model to be extracted.
        start_marker (str | Sequence[str]): Start marker(s) to extract.
        end_marker (str | Sequence[str]): End marker(s) to extract.
        start_name_map (Dict[str, str]): A mapping of start names, defaults to
            `None`.
        end_name_map (Dict[str, str]): A mapping of end names, defaults to
            `None`.
        dynamic_axes (Dict[str, Dict[int, str]]): A dictionary to specify
            dynamic axes of input/output, defaults to `None`.
        save_file (str): A file to save the extracted model, defaults to
            `None`.

    Returns:
        onnx.ModelProto: The extracted model.
    """
    if isinstance(model, str):
        model = onnx.load(model)

    num_value_info = len(model.graph.value_info)
    inputs = []
    outputs = []
    logger = get_root_logger()
    if not isinstance(start_marker, (list, tuple)):
        start_marker = [start_marker]
    for s in start_marker:
        start_name, func_id, start_type = parse_extractor_io_string(s)
        for node in model.graph.node:
            if node.op_type == "Mark":
                attr = attribute_to_dict(node.attribute)
                if (
                    attr["func"] == start_name
                    and attr["type"] == start_type
                    and attr["func_id"] == func_id
                ):
                    name = node.input[0]
                    if name not in inputs:
                        new_name = get_new_name(
                            attr, mark_name=s, name_map=start_name_map
                        )
                        rename_value(model, name, new_name)
                        if not any(
                            [
                                v_info.name == new_name
                                for v_info in model.graph.value_info
                            ]
                        ):
                            new_val_info = onnx.helper.make_tensor_value_info(
                                new_name, attr["dtype"], attr["shape"]
                            )
                            model.graph.value_info.append(new_val_info)
                        inputs.append(new_name)

    logger.info(f'inputs: {", ".join(inputs)}')

    # collect outputs
    if not isinstance(end_marker, (list, tuple)):
        end_marker = [end_marker]
    for e in end_marker:
        end_name, func_id, end_type = parse_extractor_io_string(e)
        for node in model.graph.node:
            if node.op_type == "Mark":
                attr = attribute_to_dict(node.attribute)
                if (
                    attr["func"] == end_name
                    and attr["type"] == end_type
                    and attr["func_id"] == func_id
                ):
                    name = node.output[0]
                    if name not in outputs:
                        new_name = get_new_name(
                            attr, mark_name=e, name_map=end_name_map
                        )
                        rename_value(model, name, new_name)
                        if not any(
                            [
                                v_info.name == new_name
                                for v_info in model.graph.value_info
                            ]
                        ):
                            new_val_info = onnx.helper.make_tensor_value_info(
                                new_name, attr["dtype"], attr["shape"]
                            )
                            model.graph.value_info.append(new_val_info)
                        outputs.append(new_name)

    logger.info(f'outputs: {", ".join(outputs)}')

    # replace Mark with Identity
    for node in model.graph.node:
        if node.op_type == "Mark":
            del node.attribute[:]
            node.domain = ""
            node.op_type = "Identity"

    extractor = create_extractor(model)
    extracted_model = extractor.extract_model(inputs, outputs)

    # remove all Identity, this may be done by onnx simplifier
    remove_identity(extracted_model)

    # collect all used inputs
    used = set()
    for node in extracted_model.graph.node:
        for input in node.input:
            used.add(input)

    for output in extracted_model.graph.output:
        used.add(output.name)

    # delete unused inputs
    success = True
    while success:
        success = False
        for i, input in enumerate(extracted_model.graph.input):
            if input.name not in used:
                del extracted_model.graph.input[i]
                success = True
                break

    # eliminate output without shape
    for xs in [extracted_model.graph.output]:
        for x in xs:
            if not x.type.tensor_type.shape.dim:
                logger.info(f"fixing output shape: {x.name}")
                x.CopyFrom(
                    onnx.helper.make_tensor_value_info(
                        x.name, x.type.tensor_type.elem_type, []
                    )
                )

    # eliminate 0-batch dimension, dirty workaround for two-stage detectors
    for input in extracted_model.graph.input:
        if input.name in inputs:
            if input.type.tensor_type.shape.dim[0].dim_value == 0:
                input.type.tensor_type.shape.dim[0].dim_value = 1

    # eliminate duplicated value_info for inputs
    success = True
    # num_value_info == 0 if dynamic shape
    if num_value_info == 0:
        while len(extracted_model.graph.value_info) > 0:
            extracted_model.graph.value_info.pop()
    while success:
        success = False
        for i, x in enumerate(extracted_model.graph.value_info):
            if x.name in inputs:
                del extracted_model.graph.value_info[i]
                success = True
                break

    # dynamic shape support
    if dynamic_axes is not None:
        for input_node in extracted_model.graph.input:
            if input_node.name in dynamic_axes:
                axes = dynamic_axes[input_node.name]
                for k, v in axes.items():
                    input_node.type.tensor_type.shape.dim[k].dim_value = 0
                    input_node.type.tensor_type.shape.dim[k].dim_param = v
        for output_node in extracted_model.graph.output:
            for idx, dim in enumerate(output_node.type.tensor_type.shape.dim):
                dim.dim_value = 0
                dim.dim_param = f"dim_{idx}"

    # remove mmdeploy domain if useless
    remove_imports(extracted_model)

    # save extract_model if save_file is given
    if save_file is not None:
        onnx.save(extracted_model, save_file)

    return extracted_model



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
from mmengine.config import ConfigDict
from mmengine.structures import InstanceData
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.dense_heads.fovea_head.FoveaHead.predict_by_feat"
)
def fovea_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
) -> InstanceData:
    """Rewrite `predict_by_feat` of `FoveaHead` for default backend.

    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.

    Args:
        ctx (ContextCaller): The context with additional information.
        self (FoveaHead): The instance of the class FoveaHead.
        cls_scores (list[Tensor]): Box scores for each scale level
            with shape (N, num_anchors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for each scale
            level with shape (N, num_anchors * 4, H, W).
        score_factors (list[Tensor], Optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Default None.
        batch_img_metas (list[dict]):  Meta information of the image, e.g.,
            image size, scaling factor, etc.
        cfg (mmengine.Config | None): Test / postprocessing configuration,
            if None, test_cfg would be used. Default: None.
        rescale (bool): If True, return boxes in original image space.
            Default: False.

    Returns:
        tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
    """
    ctx = FUNCTION_REWRITER.get_context()
    assert len(cls_scores) == len(bbox_preds)
    cfg = self.test_cfg if cfg is None else cfg
    num_levels = len(cls_scores)
    featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, dtype=bbox_preds[0].dtype, device=bbox_preds[0].device
    )
    cls_score_list = [cls_scores[i].detach() for i in range(num_levels)]
    bbox_pred_list = [bbox_preds[i].detach() for i in range(num_levels)]
    img_shape = batch_img_metas[0]["img_shape"]
    batch_size = cls_scores[0].shape[0]

    det_bboxes = []
    det_scores = []
    for cls_score, bbox_pred, base_len, point in zip(
        cls_score_list, bbox_pred_list, self.base_edge_list, mlvl_priors
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]
        x = point[:, 0]
        y = point[:, 1]
        scores = (
            cls_score.permute(0, 2, 3, 1)
            .reshape(batch_size, -1, self.cls_out_channels)
            .sigmoid()
        )
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, 4).exp()
        x1 = (x - base_len * bbox_pred[:, :, 0]).clamp(min=0, max=img_shape[1] - 1)
        y1 = (y - base_len * bbox_pred[:, :, 1]).clamp(min=0, max=img_shape[0] - 1)
        x2 = (x + base_len * bbox_pred[:, :, 2]).clamp(min=0, max=img_shape[1] - 1)
        y2 = (y + base_len * bbox_pred[:, :, 3]).clamp(min=0, max=img_shape[0] - 1)
        bboxes = torch.stack([x1, y1, x2, y2], -1)
        det_bboxes.append(bboxes)
        det_scores.append(scores)
    det_bboxes = torch.cat(det_bboxes, dim=1)
    if rescale:
        scale_factor = batch_img_metas["scale_factor"]
        det_bboxes /= det_bboxes.new_tensor(scale_factor)
    det_scores = torch.cat(det_scores, dim=1)

    deploy_cfg = ctx.cfg
    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    nms_pre = cfg.get("deploy_nms_pre", -1)
    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        det_bboxes,
        det_scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=nms_pre,
        keep_top_k=cfg.max_per_img,
    )



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmcv.ops import RoIAlign
from torch.autograd import Function

from juxtapose.mmdeploy.core.optimizers import mark
from juxtapose.mmdeploy.core.rewriters import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import get_backend
from juxtapose.mmdeploy.utils.constants import Backend


class MultiLevelRoiAlign(Function):
    """Create MMCVMultiLevelRoiAlign op.

    This class is used to create a MultiLevelRoiAlign in ONNX for the TensorRT
    backend.
    """

    @staticmethod
    def symbolic(g, *args):
        """Symbolic function for creating onnx op."""
        aligned = args[-1]
        featmap_strides = args[-2]
        finest_scale = args[-3]
        roi_scale_factor = args[-4]
        sampling_ratio = args[-5]
        pool_mode = args[-6]
        pool_mode_flag = 0 if pool_mode == "max" else 1
        output_size = args[-7]
        inputs = args[: len(featmap_strides)]
        rois = args[len(featmap_strides)]
        return g.op(
            "mmdeploy::MMCVMultiLevelRoiAlign",
            rois,
            *inputs,
            output_height_i=output_size[1],
            output_width_i=output_size[0],
            pool_mode_i=pool_mode_flag,
            sampling_ratio_i=sampling_ratio,
            roi_scale_factor_f=roi_scale_factor,
            finest_scale_i=finest_scale,
            featmap_strides_f=featmap_strides,
            aligned_i=aligned,
        )

    @staticmethod
    def forward(g, *args):
        """Run forward."""
        # aligned = args[-1]
        featmap_strides = args[-2]
        # finest_scale = args[-3]
        # roi_scale_factor = args[-4]
        # sampling_ratio = args[-5]
        output_size = args[-7]
        inputs = args[: len(featmap_strides)]
        rois = args[len(featmap_strides)]

        num_proposals = rois.shape[0]
        channel = inputs[0].shape[1]

        return rois.new_zeros((num_proposals, channel, output_size[1], output_size[0]))


@mark("roi_extractor", inputs=["feats", "rois"], outputs=["bbox_feats"])
@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.roi_extractors."
    "single_level_roi_extractor.SingleRoIExtractor.forward",
    backend="tensorrt",
)
def single_roi_extractor__forward__tensorrt(self, feats, rois, roi_scale_factor=None):
    """Rewrite `forward` of `SingleRoIExtractor` for TensorRT backend.

    This function uses MMCVMultiLevelRoiAlign op for TensorRT deployment.
    """
    featmap_strides = self.featmap_strides
    finest_scale = self.finest_scale

    for roi_layer in self.roi_layers:
        assert isinstance(
            roi_layer, RoIAlign
        ), f"{type(roi_layer)} is not supported in TensorRT."

    roi_layer = self.roi_layers[0]
    out_size = roi_layer.output_size
    sampling_ratio = roi_layer.sampling_ratio
    pool_mode = roi_layer.pool_mode
    aligned = roi_layer.aligned
    if roi_scale_factor is None:
        roi_scale_factor = 1.0

    featmap_strides = [float(s) for s in featmap_strides]
    return MultiLevelRoiAlign.apply(
        *feats,
        rois,
        out_size,
        pool_mode,
        sampling_ratio,
        roi_scale_factor,
        finest_scale,
        featmap_strides,
        aligned,
    )


class AscendRoiExtractor(Function):
    """Create AscendRoiExtractor op.

    This class is used to create a AscendRoiExtractor in ONNX for the Ascend
    backend.
    """

    @staticmethod
    def symbolic(g, *args):
        """Symbolic function for creating onnx op."""
        aligned = args[-1]
        featmap_strides = [1 / stride for stride in args[-2]]
        finest_scale = args[-3]
        roi_scale_factor = args[-4]
        sampling_ratio = args[-5]
        pool_mode = args[-6]
        output_size = args[-7]
        inputs = args[: len(featmap_strides)]
        rois = args[len(featmap_strides)]

        return g.op(
            "mmdeploy::RoiExtractor",
            *inputs,
            rois,
            pooled_height_i=output_size[1],
            pooled_width_i=output_size[0],
            pool_mode_s=pool_mode,
            sample_num_i=sampling_ratio,
            roi_scale_factor_f=roi_scale_factor,
            finest_scale_i=finest_scale,
            spatial_scale_f=featmap_strides,
            aligned_i=aligned,
            outputs=1,
        )

    @staticmethod
    def forward(ctx, *args):
        """Run forward."""
        # aligned = args[-1]
        featmap_strides = args[-2]
        # finest_scale = args[-3]
        # roi_scale_factor = args[-4]
        # sampling_ratio = args[-5]
        output_size = args[-7]
        inputs = args[: len(featmap_strides)]
        rois = args[len(featmap_strides)]

        num_proposals = rois.shape[0]
        channel = inputs[0].shape[1]

        return rois.new_zeros((num_proposals, channel, output_size[1], output_size[0]))


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.roi_extractors."
    "single_level_roi_extractor.SingleRoIExtractor.forward",
    backend="ascend",
)
def single_roi_extractor__forward__ascend(self, feats, rois, roi_scale_factor=None):
    """Rewrite `forward` of `SingleRoIExtractor` for Ascend backend.

    This function uses RoiExtractor op for Ascend deployment.
    """
    featmap_strides = self.featmap_strides
    finest_scale = self.finest_scale

    for roi_layer in self.roi_layers:
        assert isinstance(
            roi_layer, RoIAlign
        ), f"{type(roi_layer)} is not supported in Ascend."

    roi_layer = self.roi_layers[0]
    out_size = roi_layer.output_size
    sampling_ratio = roi_layer.sampling_ratio
    pool_mode = roi_layer.pool_mode
    aligned = roi_layer.aligned
    if roi_scale_factor is None:
        roi_scale_factor = 1.0

    featmap_strides = [float(s) for s in featmap_strides]
    return AscendRoiExtractor.apply(
        *feats,
        rois,
        out_size,
        pool_mode,
        sampling_ratio,
        roi_scale_factor,
        finest_scale,
        featmap_strides,
        aligned,
    )


@mark("roi_extractor", inputs=["feats", "rois"], outputs=["bbox_feats"])
@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.roi_heads.SingleRoIExtractor.forward"
)
def single_roi_extractor__forward(self, feats, rois, roi_scale_factor=None):
    """Rewrite `forward` of SingleRoIExtractor for default backend.

    Rewrite this function to:
    1. enable exporting to IR even though the input
    image contains no targets. Note that, `ScatterND` of onnx may conflict with
    `Reshape` if a tensor have a dim size of 0. Thus, we have to cat zeros to
    the dim 0 of `roi_feats` and recover back after all roi align finished.

    2. this function adds mark for roi_extractor forward and remove
    unnecessary code of origin forward function when using ONNX as IR.

    3. use the roi align in torhcvision to accelerate the inference.
    """
    ctx = FUNCTION_REWRITER.get_context(
        "mmdet.models.roi_heads.SingleRoIExtractor.forward"
    )
    backend = get_backend(ctx.cfg)
    out_size = self.roi_layers[0].output_size
    num_levels = len(feats)
    roi_feats = feats[0].new_zeros(rois.shape[0], self.out_channels, *out_size)
    if num_levels == 1:
        assert len(rois) > 0, "The number of rois should be positive"
        if backend == Backend.TORCHSCRIPT or backend == Backend.COREML:
            self.roi_layers[0].use_torchvision = True
        return self.roi_layers[0](feats[0], rois)

    target_lvls = self.map_roi_levels(rois, num_levels)

    if roi_scale_factor is not None:
        rois = self.roi_rescale(rois, roi_scale_factor)

    # concate zeros to rois and roi_feats for empty tensor cases
    roi_feats = torch.cat(
        (roi_feats.new_zeros(num_levels * 2, *roi_feats.shape[-3:]), roi_feats)
    )
    rois = torch.cat((rois.new_zeros(num_levels * 2, 5), rois))
    _tmp = torch.linspace(
        0,
        num_levels - 1,
        num_levels,
        dtype=target_lvls.dtype,
        device=target_lvls.device,
    )
    target_lvls = torch.cat((_tmp, _tmp, target_lvls))
    for i in range(num_levels):
        mask = target_lvls == i
        inds = mask.nonzero(as_tuple=False).squeeze(1)
        rois_t = rois[inds]
        # use the roi align in torhcvision
        if backend == Backend.TORCHSCRIPT or backend == Backend.COREML:
            self.roi_layers[i].use_torchvision = True
        roi_feats_t = self.roi_layers[i](feats[i], rois_t)
        roi_feats[inds] = roi_feats_t
    # slice to recover original size
    roi_feats = roi_feats[num_levels * 2 :]
    return roi_feats


class SingleRoIExtractorOpenVINO(Function):
    """This class adds support for ExperimentalDetectronROIFeatureExtractor
    when exporting to OpenVINO.

    The `forward` method returns the original output, which is calculated in
    advance and added to the SingleRoIExtractorOpenVINO class. In addition, the
    list of arguments is changed here to be more suitable for
    ExperimentalDetectronROIFeatureExtractor.
    """

    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def forward(g, output_size, featmap_strides, sample_num, rois, *feats):
        """Run forward."""
        return SingleRoIExtractorOpenVINO.origin_output

    @staticmethod
    def symbolic(g, output_size, featmap_strides, sample_num, rois, *feats):
        """Symbolic function for creating onnx op."""
        from torch.onnx.symbolic_opset10 import _slice

        rois = _slice(g, rois, axes=[1], starts=[1], ends=[5])
        domain = "org.openvinotoolkit"
        op_name = "ExperimentalDetectronROIFeatureExtractor"
        roi_feats = g.op(
            f"{domain}::{op_name}",
            rois,
            *feats,
            output_size_i=output_size,
            pyramid_scales_i=featmap_strides,
            sampling_ratio_i=sample_num,
            image_id_i=0,
            distribute_rois_between_levels_i=1,
            preserve_rois_order_i=0,
            aligned_i=1,
            outputs=1,
        )
        return roi_feats


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.roi_extractors."
    "single_level_roi_extractor.SingleRoIExtractor.forward",
    backend="openvino",
)
def single_roi_extractor__forward__openvino(self, feats, rois, roi_scale_factor=None):
    """Replaces SingleRoIExtractor with SingleRoIExtractorOpenVINO when
    exporting to OpenVINO.

    This function uses ExperimentalDetectronROIFeatureExtractor for OpenVINO.
    """
    ctx = FUNCTION_REWRITER.get_context()

    # Adding original output to SingleRoIExtractorOpenVINO.
    state = torch._C._get_tracing_state()
    origin_output = ctx.origin_func(self, feats, rois, roi_scale_factor)
    setattr(SingleRoIExtractorOpenVINO, "origin_output", origin_output)
    torch._C._set_tracing_state(state)

    output_size = self.roi_layers[0].output_size[0]
    featmap_strides = self.featmap_strides
    sample_num = self.roi_layers[0].sampling_ratio

    args = (output_size, featmap_strides, sample_num, rois, *feats)
    result = SingleRoIExtractorOpenVINO.apply(*args)
    return result


@mark("roi_extractor", inputs=["feats", "rois"], outputs=["bbox_feats"])
@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.roi_heads.SingleRoIExtractor.forward",
    backend=Backend.TVM.value,
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.roi_heads.SingleRoIExtractor.forward",
    backend=Backend.COREML.value,
)
def single_roi_extractor__forward__coreml(self, feats, rois, roi_scale_factor=None):
    """Rewrite `forward` of SingleRoIExtractor for coreml."""
    ctx = FUNCTION_REWRITER.get_context()
    backend = get_backend(ctx.cfg)
    out_size = self.roi_layers[0].output_size
    num_levels = len(feats)
    roi_feats = feats[0].new_zeros(rois.shape[0], self.out_channels, *out_size)
    if num_levels == 1:
        assert len(rois) > 0, "The number of rois should be positive"
        self.roi_layers[0].use_torchvision = True
        return self.roi_layers[0](feats[0], rois)

    target_lvls = self.map_roi_levels(rois, num_levels)

    if roi_scale_factor is not None:
        rois = self.roi_rescale(rois, roi_scale_factor)

    for i in range(num_levels):
        mask = target_lvls == i
        # inds = mask.nonzero(as_tuple=False).squeeze(1)
        rois_t = rois * mask.unsqueeze(-1)
        # use the roi align in torhcvision
        if backend == Backend.COREML:
            self.roi_layers[i].use_torchvision = True
        roi_feats_t = self.roi_layers[i](feats[i], rois_t)
        roi_feats = roi_feats + roi_feats_t * (rois_t[:, -1] > 0).reshape(-1, 1, 1, 1)
    # slice to recover original size
    return roi_feats



# Copyright (c) OpenMMLab. All rights reserved.
import logging
from typing import Any, Optional, Sequence

import mmengine

from juxtapose.mmdeploy.codebase import BaseTask, get_codebase_class, import_codebase
from juxtapose.mmdeploy.utils import (
    get_backend,
    get_codebase,
    get_task_type,
    parse_device_id,
)
from juxtapose.mmdeploy.utils.config_utils import get_codebase_external_module
from ..core import PIPELINE_MANAGER


def check_backend_device(deploy_cfg: mmengine.Config, device: str):
    """Check if device is appropriate for the backend.

    Args:
        deploy_cfg (str | mmengine.Config): Deployment config file.
        device (str): A string specifying device type.
    """
    backend = get_backend(deploy_cfg).value
    device_id = parse_device_id(device)
    mismatch = dict(
        tensorrt=lambda id: id == -1,
        openvino=lambda id: id > -1,
    )
    if backend in mismatch and mismatch[backend](device_id):
        raise ValueError(f"{device} is invalid for the backend {backend}")


def build_task_processor(
    model_cfg: mmengine.Config, deploy_cfg: mmengine.Config, device: str
) -> BaseTask:
    """Build a task processor to manage the deployment pipeline.

    Args:
        model_cfg (str | mmengine.Config): Model config file.
        deploy_cfg (str | mmengine.Config): Deployment config file.
        device (str): A string specifying device type.

    Returns:
        BaseTask: A task processor.
    """
    check_backend_device(deploy_cfg=deploy_cfg, device=device)
    codebase_type = get_codebase(deploy_cfg, model_cfg=model_cfg)
    custom_module_list = get_codebase_external_module(deploy_cfg)
    import_codebase(codebase_type, custom_module_list)
    codebase = get_codebase_class(codebase_type)
    return codebase.build_task_processor(model_cfg, deploy_cfg, device)


def get_predefined_partition_cfg(deploy_cfg: mmengine.Config, partition_type: str):
    """Get the predefined partition config.

    Notes:
        Currently only support mmdet codebase.

    Args:
        deploy_cfg (mmengine.Config): use deploy config to get the codebase and
            task type.
        partition_type (str): A string specifying partition type.

    Returns:
        dict: A dictionary of partition config.
    """
    codebase_type = get_codebase(deploy_cfg)
    custom_module_list = get_codebase_external_module(deploy_cfg)
    import_codebase(codebase_type, custom_module_list)
    task = get_task_type(deploy_cfg)
    codebase = get_codebase_class(codebase_type)
    task_processor_class = codebase.get_task_class(task)
    return task_processor_class.get_partition_cfg(partition_type)


@PIPELINE_MANAGER.register_pipeline()
def to_backend(
    backend_name: str,
    ir_files: Sequence[str],
    work_dir: str,
    deploy_cfg: Optional[Any] = None,
    log_level: int = logging.INFO,
    device: str = "cpu",
    **kwargs,
) -> Sequence[str]:
    """Convert intermediate representation to given backend.

    Args:
        backend_name (str): The name of the backend.
        ir_files (Sequence[str]): The intermediate representation files.
        work_dir (str): The work directory, backend files and logs should
            be save in this directory.
        deploy_cfg (Any): The deploy config.
        log_level (int, optional): The log level. Defaults to logging.INFO.
        device (str, optional): The device type. Defaults to 'cpu'.
    Returns:
        Sequence[str]: Backend files.
    """
    from juxtapose.mmdeploy.backend.base import get_backend_manager

    backend_mgr = get_backend_manager(backend_name)
    return backend_mgr.to_backend(
        ir_files=ir_files,
        work_dir=work_dir,
        deploy_cfg=deploy_cfg,
        log_level=log_level,
        device=device,
        **kwargs,
    )



# Copyright (c) OpenMMLab. All rights reserved.
from .backend_manager import ONNXRuntimeManager

_BackendManager = ONNXRuntimeManager

is_available = _BackendManager.is_available
build_wrapper = _BackendManager.build_wrapper

__all__ = ['ONNXRuntimeManager']

if is_available():
    try:
        # import wrapper if pytorch is available
        from .wrapper import ORTWrapper
        __all__ += ['ORTWrapper']
    except Exception:
        pass



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
from mmengine import ConfigDict
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import (
    gather_topk,
    get_post_processing_params,
    pad_with_value_if_necessary,
)
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import Backend, is_dynamic_shape


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.rpn_head." "RPNHead.predict_by_feat"
)
def rpn_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
    **kwargs,
):
    """Rewrite `predict_by_feat` of `RPNHead` for default backend.

    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.

    Args:
        ctx (ContextCaller): The context with additional information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        score_factors (list[Tensor], optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Defaults to None.
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.

    Returns:
        If with_nms == True:
            tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
        Else:
            tuple[Tensor, Tensor, Tensor]: batch_mlvl_bboxes,
                batch_mlvl_scores, batch_mlvl_centerness
    """
    ctx = FUNCTION_REWRITER.get_context()
    img_metas = batch_img_metas
    assert len(cls_scores) == len(bbox_preds)
    deploy_cfg = ctx.cfg
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    num_levels = len(cls_scores)

    device = cls_scores[0].device
    featmap_sizes = [cls_scores[i].shape[-2:] for i in range(num_levels)]
    mlvl_anchors = self.anchor_generator.grid_anchors(featmap_sizes, device=device)

    mlvl_cls_scores = [cls_scores[i].detach() for i in range(num_levels)]
    mlvl_bbox_preds = [bbox_preds[i].detach() for i in range(num_levels)]
    assert len(mlvl_cls_scores) == len(mlvl_bbox_preds) == len(mlvl_anchors)

    cfg = self.test_cfg if cfg is None else cfg
    batch_size = mlvl_cls_scores[0].shape[0]
    pre_topk = cfg.get("nms_pre", -1)

    # loop over features, decode boxes
    mlvl_valid_bboxes = []
    mlvl_scores = []
    mlvl_valid_anchors = []
    for level_id, cls_score, bbox_pred, anchors in zip(
        range(num_levels), mlvl_cls_scores, mlvl_bbox_preds, mlvl_anchors
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]
        cls_score = cls_score.permute(0, 2, 3, 1)
        if self.use_sigmoid_cls:
            cls_score = cls_score.reshape(batch_size, -1)
            scores = cls_score.sigmoid()
        else:
            cls_score = cls_score.reshape(batch_size, -1, 2)
            # We set FG labels to [0, num_class-1] and BG label to
            # num_class in RPN head since mmdet v2.5, which is unified to
            # be consistent with other head since mmdet v2.0. In mmdet v2.0
            # to v2.4 we keep BG label as 0 and FG label as 1 in rpn head.
            scores = cls_score.softmax(-1)[..., 0]
        scores = scores.reshape(batch_size, -1, 1)
        dim = self.bbox_coder.encode_size
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, dim)

        # use static anchor if input shape is static
        if not is_dynamic_flag:
            anchors = anchors.data

        anchors = anchors.unsqueeze(0)

        # topk in tensorrt does not support shape<k
        # concate zero to enable topk,
        scores = pad_with_value_if_necessary(scores, 1, pre_topk, 0.0)
        bbox_pred = pad_with_value_if_necessary(bbox_pred, 1, pre_topk)
        anchors = pad_with_value_if_necessary(anchors, 1, pre_topk)

        if pre_topk > 0:
            _, topk_inds = scores.squeeze(2).topk(pre_topk)
            bbox_pred, scores = gather_topk(
                bbox_pred,
                scores,
                inds=topk_inds,
                batch_size=batch_size,
                is_batched=True,
            )
            anchors = gather_topk(
                anchors, inds=topk_inds, batch_size=batch_size, is_batched=False
            )
        mlvl_valid_bboxes.append(bbox_pred)
        mlvl_scores.append(scores)
        mlvl_valid_anchors.append(anchors)

    batch_mlvl_bboxes = torch.cat(mlvl_valid_bboxes, dim=1)
    batch_mlvl_scores = torch.cat(mlvl_scores, dim=1)
    batch_mlvl_anchors = torch.cat(mlvl_valid_anchors, dim=1)
    batch_mlvl_bboxes = self.bbox_coder.decode(
        batch_mlvl_anchors, batch_mlvl_bboxes, max_shape=img_metas[0]["img_shape"]
    )
    # ignore background class
    if not self.use_sigmoid_cls:
        batch_mlvl_scores = batch_mlvl_scores[..., : self.num_classes]
    if not with_nms:
        return batch_mlvl_bboxes, batch_mlvl_scores

    post_params = get_post_processing_params(deploy_cfg)
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    # only one class in rpn
    max_output_boxes_per_class = keep_top_k
    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        batch_mlvl_bboxes,
        batch_mlvl_scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )


# TODO: Fix for 1.x
@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.dense_heads.RPNHead.get_bboxes", backend=Backend.NCNN.value
)
def rpn_head__get_bboxes__ncnn(
    self, cls_scores, bbox_preds, img_metas, with_nms=True, cfg=None, **kwargs
):
    """Rewrite `get_bboxes` of `RPNHead` for ncnn backend.

    Shape node and batch inference is not supported by ncnn. This function
    transform dynamic shape to constant shape and remove batch inference.

    Args:
        ctx (ContextCaller): The context with additional information.
        cls_scores (list[Tensor]): Box scores for each level in the
            feature pyramid, has shape
            (N, num_anchors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for each
            level in the feature pyramid, has shape
            (N, num_anchors * 4, H, W).
        img_metas (list[dict]): Meta information of each image, e.g.,
            image size, scaling factor, etc.
        with_nms (bool): If True, do nms before return boxes.
            Default: True.
        cfg (mmengine.Config | None): Test / postprocessing configuration,
            if None, test_cfg would be used.
            Default: None.


    Returns:
        If with_nms == True:
            tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
        Else:
            tuple[Tensor, Tensor]: batch_mlvl_bboxes, batch_mlvl_scores
    """
    ctx = FUNCTION_REWRITER.get_context()
    assert len(cls_scores) == len(bbox_preds)
    deploy_cfg = ctx.cfg
    assert not is_dynamic_shape(deploy_cfg)
    num_levels = len(cls_scores)

    device = cls_scores[0].device
    featmap_sizes = [cls_scores[i].shape[-2:] for i in range(num_levels)]
    mlvl_anchors = self.anchor_generator.grid_anchors(featmap_sizes, device=device)

    mlvl_cls_scores = [cls_scores[i].detach() for i in range(num_levels)]
    mlvl_bbox_preds = [bbox_preds[i].detach() for i in range(num_levels)]
    assert len(mlvl_cls_scores) == len(mlvl_bbox_preds) == len(mlvl_anchors)

    cfg = self.test_cfg if cfg is None else cfg
    batch_size = 1
    pre_topk = cfg.get("nms_pre", -1)

    # loop over features, decode boxes
    mlvl_valid_bboxes = []
    mlvl_scores = []
    mlvl_valid_anchors = []
    for level_id, cls_score, bbox_pred, anchors in zip(
        range(num_levels), mlvl_cls_scores, mlvl_bbox_preds, mlvl_anchors
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]
        cls_score = cls_score.permute(0, 2, 3, 1)
        if self.use_sigmoid_cls:
            cls_score = cls_score.reshape(batch_size, -1)
            scores = cls_score.sigmoid()
        else:
            cls_score = cls_score.reshape(batch_size, -1, 2)
            # We set FG labels to [0, num_class-1] and BG label to
            # num_class in RPN head since mmdet v2.5, which is unified to
            # be consistent with other head since mmdet v2.0. In mmdet v2.0
            # to v2.4 we keep BG label as 0 and FG label as 1 in rpn head.
            scores = cls_score.softmax(-1)[..., 0]
        scores = scores.reshape(batch_size, -1, 1)
        dim = self.bbox_coder.encode_size
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, dim)
        anchors = anchors.expand_as(bbox_pred).data

        if pre_topk > 0:
            _, topk_inds = scores.squeeze(2).topk(pre_topk)
            topk_inds = topk_inds.view(-1)
            anchors = anchors[:, topk_inds, :]
            bbox_pred = bbox_pred[:, topk_inds, :]
            scores = scores[:, topk_inds, :]

        mlvl_valid_bboxes.append(bbox_pred)
        mlvl_scores.append(scores)
        mlvl_valid_anchors.append(anchors)

    batch_mlvl_bboxes = torch.cat(mlvl_valid_bboxes, dim=1)
    batch_mlvl_scores = torch.cat(mlvl_scores, dim=1)
    batch_mlvl_anchors = torch.cat(mlvl_valid_anchors, dim=1)
    batch_mlvl_bboxes = self.bbox_coder.decode(
        batch_mlvl_anchors, batch_mlvl_bboxes, max_shape=img_metas[0]["img_shape"]
    )
    # ignore background class
    if not self.use_sigmoid_cls:
        batch_mlvl_scores = batch_mlvl_scores[..., : self.num_classes]
    if not with_nms:
        return batch_mlvl_bboxes, batch_mlvl_scores

    post_params = get_post_processing_params(deploy_cfg)
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    # only one class in rpn
    max_output_boxes_per_class = keep_top_k
    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        batch_mlvl_bboxes,
        batch_mlvl_scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )



# Copyright (c) OpenMMLab. All rights reserved.

import copy
import os
from collections import defaultdict
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import mmcv
import mmengine
import numpy as np
import torch
from mmengine.model import BaseDataPreprocessor
from mmengine.registry import Registry

from juxtapose.mmdeploy.codebase.base import CODEBASE, BaseTask, MMCodebase
from juxtapose.mmdeploy.utils import Codebase, Task, get_input_shape, get_root_logger


def process_model_config(
    model_cfg: mmengine.Config,
    imgs: Union[Sequence[str], Sequence[np.ndarray]],
    input_shape: Optional[Sequence[int]] = None,
):
    """Process the model config for sdk model.

    Args:
        model_cfg (mmengine.Config): The model config.
        imgs (Sequence[str] | Sequence[np.ndarray]): Input image(s), accepted
            data type are List[str], List[np.ndarray].
        input_shape (list[int]): A list of two integer in (width, height)
            format specifying input shape. Default: None.

    Returns:
        mmengine.Config: the model config after processing.
    """
    cfg = copy.deepcopy(model_cfg)
    test_pipeline = cfg.test_dataloader.dataset.pipeline
    data_preprocessor = cfg.model.data_preprocessor
    codec = cfg.codec
    if isinstance(codec, list):
        codec = codec[-1]
    input_size = codec["input_size"] if input_shape is None else input_shape
    test_pipeline[0] = dict(type="LoadImageFromFile")
    for i in reversed(range(len(test_pipeline))):
        trans = test_pipeline[i]
        if trans["type"] == "PackPoseInputs":
            test_pipeline.pop(i)
        elif trans["type"] == "GetBBoxCenterScale":
            trans["type"] = "TopDownGetBboxCenterScale"
            trans["padding"] = 1.25  # default argument
            trans["image_size"] = input_size
        elif trans["type"] == "TopdownAffine":
            trans["type"] = "TopDownAffine"
            trans["image_size"] = input_size
            trans.pop("input_size")

    test_pipeline.append(
        dict(
            type="Normalize",
            mean=data_preprocessor.mean,
            std=data_preprocessor.std,
            to_rgb=data_preprocessor.bgr_to_rgb,
        )
    )
    test_pipeline.append(dict(type="ImageToTensor", keys=["img"]))
    test_pipeline.append(
        dict(
            type="Collect",
            keys=["img"],
            meta_keys=[
                "img_shape",
                "pad_shape",
                "ori_shape",
                "img_norm_cfg",
                "scale_factor",
                "bbox_score",
                "center",
                "scale",
            ],
        )
    )

    cfg.test_dataloader.dataset.pipeline = test_pipeline
    return cfg


def _get_dataset_metainfo(model_cfg: mmengine.Config):
    """Get metainfo of dataset.

    Args:
        model_cfg Config: Input model Config object.
    Returns:
        (list[str], list[np.ndarray]): Class names and palette
    """
    from mmpose import datasets  # noqa
    from mmpose.registry import DATASETS

    module_dict = DATASETS.module_dict

    for dataloader_name in ["test_dataloader", "val_dataloader", "train_dataloader"]:
        if dataloader_name not in model_cfg:
            continue
        dataloader_cfg = model_cfg[dataloader_name]
        dataset_cfg = dataloader_cfg.dataset
        dataset_mmpose = module_dict.get(dataset_cfg.type, None)
        if dataset_mmpose is None:
            continue
        if hasattr(dataset_mmpose, "_load_metainfo") and isinstance(
            dataset_mmpose._load_metainfo, Callable
        ):
            meta = dataset_mmpose._load_metainfo(dataset_cfg.get("metainfo", None))
            if meta is not None:
                return meta
        if hasattr(dataset_mmpose, "METAINFO"):
            return dataset_mmpose.METAINFO

    return None


MMPOSE_TASK = Registry("mmpose_tasks")


@CODEBASE.register_module(Codebase.MMPOSE.value)
class MMPose(MMCodebase):
    """mmpose codebase class."""

    task_registry = MMPOSE_TASK

    @classmethod
    def register_deploy_modules(cls):
        """register rewritings."""
        import juxtapose.mmdeploy.codebase.mmdet.models
        import juxtapose.mmdeploy.codebase.mmdet.ops
        import juxtapose.mmdeploy.codebase.mmdet.structures
        import juxtapose.mmdeploy.codebase.mmpose.models  # noqa: F401

    @classmethod
    def register_all_modules(cls):
        """register all modules from mmpose."""
        from mmpose.utils.setup_env import register_all_modules

        cls.register_deploy_modules()
        register_all_modules(True)


@MMPOSE_TASK.register_module(Task.POSE_DETECTION.value)
class PoseDetection(BaseTask):
    """Pose detection task class.

    Args:
        model_cfg (mmengine.Config): Original PyTorch model config file.
        deploy_cfg (mmengine.Config): Deployment config file or loaded Config
            object.
        device (str): A string represents device type.
    """

    def __init__(
        self, model_cfg: mmengine.Config, deploy_cfg: mmengine.Config, device: str
    ):
        super().__init__(model_cfg, deploy_cfg, device)
        self.model_cfg.model.test_cfg["flip_test"] = False

        from mmcv.transforms import Compose
        from mmpose.registry import TRANSFORMS

        cfg = self.model_cfg

        # build the data pipeline
        test_pipeline = [
            TRANSFORMS.build(c) for c in cfg.test_dataloader.dataset.pipeline
        ]
        self.test_pipeline = Compose(test_pipeline)

    def build_backend_model(
        self,
        model_files: Sequence[str] = None,
        data_preprocessor_updater: Optional[Callable] = None,
        **kwargs,
    ) -> torch.nn.Module:
        """build backend model.

        Args:
            model_files (Sequence[str]): Input model files. Default is None.
            data_preprocessor_updater (Callable | None): A function to update
                the data_preprocessor. Defaults to None.

        Returns:
            nn.Module: An initialized backend model.
        """
        from .pose_detection_model import build_pose_detection_model

        data_preprocessor = self.model_cfg.model.data_preprocessor
        if data_preprocessor_updater is not None:
            data_preprocessor = data_preprocessor_updater(data_preprocessor)
        model = build_pose_detection_model(
            model_files,
            self.model_cfg,
            self.deploy_cfg,
            device=self.device,
            data_preprocessor=data_preprocessor,
            **kwargs,
        )
        return model.eval().to(self.device)

    def create_input(
        self,
        imgs: Union[str, np.ndarray, Sequence],
        input_shape: Sequence[int] = None,
        data_preprocessor: Optional[BaseDataPreprocessor] = None,
        bboxes: Sequence = None,
        **kwargs,
    ) -> Tuple[Dict, torch.Tensor]:
        """Create input for pose detection.

        Args:
            imgs (Any): Input image(s), accepted data type are ``str``,
                ``np.ndarray``.
            input_shape (list[int]): A list of two integer in (width, height)
                format specifying input shape. Defaults to ``None``.
            data_preprocessor (BaseDataPreprocessor | None): Input data pre-
                processor. Default is ``None``.

        Returns:
            tuple: (data, inputs), meta information for the input image
             and input.
        """
        from mmcv.transforms import Compose
        from mmpose.registry import TRANSFORMS

        cfg = self.model_cfg
        if isinstance(imgs, (list, tuple)):
            if not isinstance(imgs[0], (np.ndarray, str)):
                raise AssertionError("imgs must be strings or numpy arrays")
        elif isinstance(imgs, (np.ndarray, str)):
            imgs = [imgs]
            img_path = [imgs]
        else:
            raise AssertionError("imgs must be strings or numpy arrays")
        if isinstance(imgs, (list, tuple)) and isinstance(imgs[0], str):
            img_path = imgs
            img_data = [mmcv.imread(img) for img in imgs]
            imgs = img_data

        if not bboxes:
            person_results = []
            bboxes = []
            for img in imgs:
                height, width = img.shape[:2]
                # create dummy person results
                person_results.append([{"bbox": np.array([0, 0, width, height])}])
                bboxes.append(np.array([box["bbox"] for box in person_results[-1]]))

        if input_shape is not None and hasattr(cfg, "codec"):
            if isinstance(cfg.codec, dict):
                codec = cfg.codec
            elif isinstance(cfg.codec, list):
                codec = cfg.codec[0]
            else:
                raise TypeError(f"Unsupported type {type(cfg.codec)}")
            input_size = codec["input_size"]
            if tuple(input_shape) != tuple(input_size):
                logger = get_root_logger()
                logger.warning(
                    f"Input shape from deploy config is not "
                    f"same as input_size in model config:"
                    f"{input_shape} vs {input_size}"
                )
        batch_data = defaultdict(list)
        meta_data = _get_dataset_metainfo(self.model_cfg)

        if bboxes:
            for i in range(len(bboxes)):
                for bbox in bboxes[i]:
                    # prepare data
                    bbox_score = np.array(
                        [bbox[4] if len(bbox) == 5 else 1]
                    )  # shape (1,)
                    data = {
                        "img": imgs[0],
                        "bbox_score": bbox_score,
                        "bbox": []
                        if hasattr(cfg.model, "bbox_head")
                        and cfg.model.bbox_head.type == "YOLOXPoseHead"
                        else bbox[None],
                        "img_path": img_path[0],
                    }
                    data.update(meta_data)
                    data = self.test_pipeline(data)
                    data["inputs"] = data["inputs"].to(self.device)
                    batch_data["inputs"].append(data["inputs"])
                    batch_data["data_samples"].append(data["data_samples"])

        else:
            assert len(imgs) == len(bboxes) == len(person_results)
            for i in range(len(imgs)):
                for bbox in bboxes[i]:
                    # prepare data
                    bbox_score = np.array(
                        [bbox[4] if len(bbox) == 5 else 1]
                    )  # shape (1,)
                    data = {
                        "img": imgs[i],
                        "bbox_score": bbox_score,
                        "bbox": []
                        if hasattr(cfg.model, "bbox_head")
                        and cfg.model.bbox_head.type == "YOLOXPoseHead"
                        else bbox[None],
                        "img_path": img_path[i],
                    }
                    data.update(meta_data)
                    data = self.test_pipeline(data)
                    data["inputs"] = data["inputs"].to(self.device)
                    batch_data["inputs"].append(data["inputs"])
                    batch_data["data_samples"].append(data["data_samples"])

        if data_preprocessor is not None:
            batch_data = data_preprocessor(batch_data, False)
        input_tensor = batch_data["inputs"]
        return batch_data, input_tensor

    def visualize(
        self,
        image: Union[str, np.ndarray],
        result: list,
        output_file: str,
        window_name: str,
        show_result: bool = False,
        **kwargs,
    ):
        """Visualize predictions of a model.

        Args:
            image (str | np.ndarray): Input image to draw predictions on.
            result (list): A list of predictions.
            output_file (str): Output file to save drawn image.
            window_name (str): The name of visualization window. Defaults to
                an empty string.
            show_result (bool): Whether to show result in windows, defaults
                to `False`.
        """
        from mmpose.apis.inference import dataset_meta_from_config
        from mmpose.visualization import PoseLocalVisualizer

        save_dir, filename = os.path.split(output_file)
        name = os.path.splitext(filename)[0]
        dataset_meta = dataset_meta_from_config(self.model_cfg, dataset_mode="test")
        visualizer = PoseLocalVisualizer(name=name, save_dir=save_dir)
        visualizer.set_dataset_meta(dataset_meta)

        if isinstance(image, str):
            image = mmcv.imread(image, channel_order="rgb")
        draw_bbox = result.pred_instances.bboxes is not None
        if draw_bbox and isinstance(result.pred_instances.bboxes, torch.Tensor):
            result.pred_instances.bboxes = result.pred_instances.bboxes.cpu().numpy()
        visualizer.add_datasample(
            name,
            image,
            data_sample=result,
            draw_gt=False,
            draw_bbox=draw_bbox,
            show=show_result,
            out_file=output_file,
        )

    def get_model_name(self, *args, **kwargs) -> str:
        """Get the model name.

        Return:
            str: the name of the model.
        """
        assert "type" in self.model_cfg.model, "model config contains no type"
        name = self.model_cfg.model.type.lower()
        return name

    @staticmethod
    def get_partition_cfg(partition_type: str, **kwargs) -> Dict:
        """Get a certain partition config for mmpose.

        Args:
            partition_type (str): A string specifying partition type.
        """
        raise NotImplementedError("Not supported yet.")

    def get_preprocess(self, *args, **kwargs) -> Dict:
        """Get the preprocess information for SDK.

        Return:
            dict: Composed of the preprocess information.
        """
        input_shape = get_input_shape(self.deploy_cfg)
        model_cfg = process_model_config(self.model_cfg, [""], input_shape)
        preprocess = model_cfg.test_dataloader.dataset.pipeline
        return preprocess

    def get_postprocess(self, *args, **kwargs) -> Dict:
        """Get the postprocess information for SDK."""
        codec = self.model_cfg.codec
        if isinstance(codec, (list, tuple)):
            codec = codec[-1]
        component = "UNKNOWN"
        params = copy.deepcopy(self.model_cfg.model.test_cfg)
        params.update(codec)
        if self.model_cfg.model.type == "TopdownPoseEstimator":
            component = "TopdownHeatmapSimpleHeadDecode"
            if codec.type == "MSRAHeatmap":
                params["post_process"] = "default"
            elif codec.type == "UDPHeatmap":
                params["post_process"] = "default"
                params["use_udp"] = True
            elif codec.type == "MegviiHeatmap":
                params["post_process"] = "megvii"
                params["modulate_kernel"] = self.model_cfg.kernel_sizes[-1]
            elif codec.type == "SimCCLabel":
                component = "SimCCLabelDecode"
            elif codec.type == "RegressionLabel":
                component = "DeepposeRegressionHeadDecode"
            elif codec.type == "IntegralRegressionLabel":
                component = "DeepposeRegressionHeadDecode"
            else:
                raise RuntimeError(f"Unsupported codecs type: {codec.type}")
        postprocess = dict(params=params, type=component)
        return postprocess



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.tensorrt import is_available
from ..core import PIPELINE_MANAGER

__all__ = ["is_available"]

if is_available():
    from juxtapose.mmdeploy.backend.tensorrt import from_onnx as _from_onnx
    from juxtapose.mmdeploy.backend.tensorrt import load, save

    from_onnx = PIPELINE_MANAGER.register_pipeline()(_from_onnx)
    __all__ += ["from_onnx", "save", "load"]
    try:
        from juxtapose.mmdeploy.backend.tensorrt.onnx2tensorrt import (
            onnx2tensorrt as _onnx2tensorrt,
        )

        onnx2tensorrt = PIPELINE_MANAGER.register_pipeline()(_onnx2tensorrt)
        __all__ += ["onnx2tensorrt"]
    except Exception:
        pass



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import mmengine

from .constants import Backend, Codebase, Task
from .utils import deprecate, get_root_logger
from pathlib import Path


def load_config(*args) -> List[mmengine.Config]:
    """Load the configuration and check the validity.

    Args:
        args (str | Sequence[str]): The path to the config file(s).

    Returns:
        List[mmengine.Config | dict]: The content of config.
    """

    def _load_config(cfg):
        if isinstance(cfg, (str, Path)):
            cfg = mmengine.Config.fromfile(
                cfg,
                # @custom update
                # lazy_import=False,
                # use_environment_variables=False,
                # use_predefined_variables=False,
                # import_custom_modules=False,
            )
        if not isinstance(cfg, (dict, mmengine.Config)):
            raise TypeError(
                "deploy_cfg must be a filename or Config object, "
                f"but got {type(cfg)}"
            )
        return cfg

    assert len(args) > 0
    configs = [_load_config(cfg) for cfg in args]

    return configs


def get_codebase_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Get the codebase_config from the config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Dict : codebase config dict.
    """
    deploy_cfg = load_config(deploy_cfg)[0]
    codebase_config = deploy_cfg.get("codebase_config", {})
    return codebase_config


def get_task_type(deploy_cfg: Union[str, mmengine.Config]) -> Task:
    """Get the task type of the algorithm.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Task : An enumeration denotes the task type.
    """

    codebase_config = get_codebase_config(deploy_cfg)
    assert "task" in codebase_config, (
        "The codebase config of deploy config" 'requires a "task" field'
    )
    task = codebase_config["task"]
    return Task.get(task)


def register_codebase(codebase: str) -> Codebase:
    """Register a new codebase which is not included in Codebase.

    Args:
        codebase (str): The codebase name.

    Returns:
        Codebase : An enumeration denotes the codebase type.
    """
    from aenum import extend_enum

    try:
        Codebase.get(codebase)
    except Exception as e:
        logger = get_root_logger()
        extend_enum(Codebase, codebase.upper(), codebase)
        logger.warn(
            f"Failed to get codebase, got: {e}. Then export "
            f"a new codebase in Codebase {codebase.upper()}: "
            f"{codebase}"
        )
    return Codebase.get(codebase)


def get_codebase(deploy_cfg: Union[str, mmengine.Config], model_cfg=None) -> Codebase:
    """Get the codebase from the config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Codebase : An enumeration denotes the codebase type.
    """
    if model_cfg is not None:
        # using mmrazor codebase if the model is a mmrazor model.
        model_cfg: dict = model_cfg["model"]
        if model_cfg.get("_scope_", None) == "mmrazor" or model_cfg["type"].startswith(
            "mmrazor."
        ):
            return register_codebase("mmrazor")
    codebase_config = get_codebase_config(deploy_cfg)
    assert "type" in codebase_config, (
        "The codebase config of deploy config" 'requires a "type" field'
    )
    codebase = codebase_config["type"]
    return register_codebase(codebase)


def get_backend_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Get the backend_config from the config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Dict : backend config dict.
    """
    deploy_cfg = load_config(deploy_cfg)[0]
    backend_config = deploy_cfg.get("backend_config", {})
    return backend_config


def get_backend(deploy_cfg: Union[str, mmengine.Config]) -> Backend:
    """Get the backend from the config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Backend: An enumeration denotes the backend type.
    """
    backend_config = get_backend_config(deploy_cfg)
    assert "type" in backend_config, (
        "The backend config of deploy config" 'requires a "type" field'
    )
    backend = backend_config["type"]
    return Backend.get(backend)


def get_ir_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Get the IR parameters in export() from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Dict: The config dictionary of IR parameters
    """

    deploy_cfg = load_config(deploy_cfg)[0]
    ir_config = deploy_cfg.get("ir_config", None)
    if ir_config is None:
        # TODO: deprecate in future
        ir_config = deploy_cfg.get("onnx_config", {})
    return ir_config


@deprecate(dst_obj=get_ir_config)
def get_onnx_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Get the onnx parameters in export() from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        Dict: The config dictionary of onnx parameters
    """
    onnx_config = get_ir_config(deploy_cfg=deploy_cfg)
    ir_type = onnx_config.get("type", None)
    assert ir_type is None or ir_type == "onnx", (
        "Expect IR type is ONNX," f"but get {ir_type}"
    )
    return onnx_config


def is_dynamic_batch(
    deploy_cfg: Union[str, mmengine.Config], input_name: Optional[str] = None
) -> bool:
    """Check if input batch is dynamic.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.
        input_name (Optional[str]): The name of input in onnx export parameter.

    Returns:
        bool: Is config set dynamic batch (axis 0).
    """

    deploy_cfg = load_config(deploy_cfg)[0]
    ir_config = get_ir_config(deploy_cfg)

    # check if input name is in the config
    input_names = ir_config.get("input_names", None)
    if input_name is None:
        input_name = input_names[0] if input_names else "input"

    # check if dynamic axes exist
    # TODO: update this when we have more IR
    dynamic_axes = get_dynamic_axes(deploy_cfg)
    if dynamic_axes is None:
        return False

    # check if given input name exist
    input_axes = dynamic_axes.get(input_name, None)
    if input_axes is None:
        return False

    # check if 0 (batch) in input axes
    if 0 in input_axes:
        return True

    return False


def is_dynamic_shape(
    deploy_cfg: Union[str, mmengine.Config], input_name: Optional[str] = None
) -> bool:
    """Check if input shape is dynamic.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.
        input_name (Optional[str]): The name of input in onnx export parameter.

    Returns:
        bool: Is config set dynamic shape (axis 2 and 3).
    """

    # Always dynamic for exporting torchscript
    if get_backend(deploy_cfg) == Backend.TORCHSCRIPT:
        return True

    deploy_cfg = load_config(deploy_cfg)[0]
    ir_config = get_ir_config(deploy_cfg)

    # check if input name is in the config
    input_names = ir_config.get("input_names", None)
    if input_name is None:
        input_name = input_names[0] if input_names else "input"

    # check if dynamic axes exist
    # TODO: update this when we have more IR
    dynamic_axes = get_dynamic_axes(deploy_cfg)
    if dynamic_axes is None:
        return False

    # check if given input name exist
    input_axes = dynamic_axes.get(input_name, None)
    if input_axes is None:
        return False

    # check if 2 (height) and 3 (width) in input axes
    if 2 in input_axes and 3 in input_axes:
        return True

    return False


def get_input_shape(deploy_cfg: Union[str, mmengine.Config]) -> List[int]:
    """Get the input shape for static exporting.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        List[int]: The input shape for backend model (axis 2 and 3),
            e.g [512, 512].
    """
    input_shape = get_ir_config(deploy_cfg).get("input_shape", None)
    if input_shape is not None:
        assert len(input_shape) == 2, "length of input_shape should equal to 2"
    return input_shape


def cfg_apply_marks(deploy_cfg: Union[str, mmengine.Config]) -> Optional[bool]:
    """Check if the model needs to be partitioned by checking if the config
    contains 'apply_marks'.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        bool or None: Whether config contains 'apply_marks'.
    """
    partition_config = deploy_cfg.get("partition_config", None)
    if partition_config is None:
        return None

    apply_marks = partition_config.get("apply_marks", False)
    return apply_marks


def get_partition_config(deploy_cfg: Union[str, mmengine.Config]) -> Optional[Dict]:
    """Check if the model needs to be partitioned and get the config of
    partition.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        dict or None: The config of partition.
    """
    partition_config = deploy_cfg.get("partition_config", None)
    if partition_config is None:
        return None

    apply_marks = partition_config.get("apply_marks", False)
    if not apply_marks:
        return None

    return partition_config


def get_calib_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Check if the model has calibration configs.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        dict: The config of calibration.
    """

    calib_config = deploy_cfg.get("calib_config", None)
    return calib_config


def get_calib_filename(deploy_cfg: Union[str, mmengine.Config]) -> Optional[str]:
    """Check if the model needs to create calib and get filename of calib.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        str | None: Could be the filename of output calib file or None.
    """

    calib_config = get_calib_config(deploy_cfg)
    if calib_config is None:
        return None
    create_calib = calib_config.get("create_calib", False)
    if create_calib:
        calib_filename = calib_config.get("calib_file", "calib_file.h5")
        return calib_filename
    else:
        return None


def get_common_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Get common parameters from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        dict: A dict of common parameters for a model.
    """
    backend_config = deploy_cfg["backend_config"]
    model_params = backend_config.get("common_config", dict())
    return model_params


def get_quantization_config(deploy_cfg: Union[str, mmengine.Config]) -> Dict:
    """Get quantization parameters from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        dict: A dict of quantization parameters for a model.
    """
    backend_config = deploy_cfg["backend_config"]
    model_params = backend_config.get("quantization_config", dict())
    return model_params


def get_model_inputs(deploy_cfg: Union[str, mmengine.Config]) -> List[Dict]:
    """Get model input parameters from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        list[dict]: A list of dict containing input parameters for a model.
    """
    backend_config = deploy_cfg["backend_config"]
    model_params = backend_config.get("model_inputs", [])
    return model_params


def get_dynamic_axes(
    deploy_cfg: Union[str, mmengine.Config], axes_names: List[str] = None
) -> Dict[str, Union[List[int], Dict[int, str]]]:
    """Get model dynamic axes from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.
        axes_names (List[str]): List with names for dynamic axes.

    Returns:
        Dict[str, Union[List[int], Dict[int, str]]]:
            Dictionary with dynamic axes.
    """
    deploy_cfg = load_config(deploy_cfg)[0]
    ir_config = get_ir_config(deploy_cfg)

    # TODO onnx will be deprecated in the future
    onnx_config = deploy_cfg.get("onnx_config", None)
    if onnx_config is None and ir_config == {}:
        raise KeyError("Field 'onnx_config' was not found in 'deploy_cfg'.")
    dynamic_axes = ir_config.get("dynamic_axes", None)
    if dynamic_axes and not isinstance(dynamic_axes, Dict):
        if axes_names is None:
            axes_names = []
            input_names = ir_config.get("input_names", None)
            if input_names:
                axes_names += input_names
            output_names = ir_config.get("output_names", None)
            if output_names:
                axes_names += output_names
            if not axes_names:
                raise KeyError("No names were found to define dynamic axes.")
        dynamic_axes = dict(zip(axes_names, dynamic_axes))
    return dynamic_axes


def get_precision(deploy_cfg: Union[str, mmengine.Config]) -> str:
    """Get precision of config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        str: The precision of target backend.
    """
    precision = "FP32"
    deploy_cfg = load_config(deploy_cfg)[0]
    backend = get_backend(deploy_cfg=deploy_cfg)
    if backend == Backend.TENSORRT:
        common_cfg = get_common_config(deploy_cfg)
        if common_cfg.get("fp16_mode", False):
            precision = "FP16"
        if common_cfg.get("int8_mode", False):
            precision = "INT8"
    if backend == Backend.NCNN and "precision" in deploy_cfg["backend_config"]:
        precision = deploy_cfg["backend_config"]["precision"]
    return precision


def get_codebase_external_module(deploy_cfg: Union[str, mmengine.Config]) -> List:
    return get_codebase_config(deploy_cfg).get("module", [])


def get_normalization(model_cfg: Union[str, mmengine.Config]):
    """Get the Normalize transform from model config.

    Args:
        model_cfg (mmengine.Config): The content of config.
    Returns:
        dict: The Normalize transform.
    """
    model_cfg = load_config(model_cfg)[0]
    data_preprocessor = model_cfg.get("data_preprocessor", {})
    data_preprocessor.update(model_cfg.model.get("data_preprocessor", {}))
    return data_preprocessor


def get_rknn_quantization(deploy_cfg: mmengine.Config):
    """Get the flag of `do_quantization` for rknn backend.

    Args:
        deploy_cfg (mmengine.Config): The content of config.
    Returns:
        bool: Do quantization or not.
    """
    if get_backend(deploy_cfg) == Backend.RKNN:
        return get_backend_config(deploy_cfg)["quantization_config"]["do_quantization"]
    return False



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.codebase.mmdet.deploy import clip_bboxes
from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.structures.bbox.transforms.distance2bbox"  # noqa
)
def distance2bbox__default(points, distance, max_shape=None):
    """Rewrite `mmdet.core.bbox.transforms.distance2bbox`

    Decode distance prediction to bounding box.

    Args:
        ctx (ContextCaller): The context with additional information.
        points (Tensor): Shape (B, N, 2) or (N, 2).
        distance (Tensor): Distance from the given point to 4
            boundaries (left, top, right, bottom). Shape (B, N, 4) or (N, 4)
        max_shape (Sequence[int] or torch.Tensor or Sequence[
            Sequence[int]],optional): Maximum bounds for boxes, specifies
            (H, W, C) or (H, W). If priors shape is (B, N, 4), then
            the max_shape should be a Sequence[Sequence[int]]
            and the length of max_shape should also be B.

    Returns:
        Tensor: Boxes with shape (N, 4) or (B, N, 4)
    """
    x1 = points[..., 0] - distance[..., 0]
    y1 = points[..., 1] - distance[..., 1]
    x2 = points[..., 0] + distance[..., 2]
    y2 = points[..., 1] + distance[..., 3]

    bboxes = torch.stack([x1, y1, x2, y2], -1)

    if max_shape is not None:
        # clip bboxes with dynamic `min` and `max`
        x1, y1, x2, y2 = clip_bboxes(x1, y1, x2, y2, max_shape)
        bboxes = torch.stack([x1, y1, x2, y2], dim=-1)
        return bboxes

    return bboxes



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
import torch.nn.functional as F
from mmengine.config import ConfigDict
from mmengine.structures import InstanceData
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import (
    gather_topk,
    get_post_processing_params,
    pad_with_value,
)
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import Backend, get_backend, is_dynamic_shape


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.gfl_head." "GFLHead.predict_by_feat"
)
def gfl_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
) -> InstanceData:
    """Rewrite `predict_by_feat` of `GFLHead` for default backend.

    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.

    Args:
        ctx (ContextCaller): The context with additional information.
        self (FoveaHead): The instance of the class FoveaHead.
        cls_scores (list[Tensor]): Box scores for each scale level
            with shape (N, num_anchors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for each scale
            level with shape (N, num_anchors * 4, H, W).
        score_factors (list[Tensor], Optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Default None.
        batch_img_metas (list[dict]):  Meta information of the image, e.g.,
            image size, scaling factor, etc.
        cfg (Config | None): Test / postprocessing configuration,
            if None, test_cfg would be used. Default: None.
        rescale (bool): If True, return boxes in original image space.
            Default: False.

    Returns:
        If with_nms == True:
            tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
        Else:
            tuple[Tensor, Tensor, Tensor]: batch_mlvl_bboxes,
                batch_mlvl_scores, batch_mlvl_centerness
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    backend = get_backend(deploy_cfg)
    num_levels = len(cls_scores)

    featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, dtype=bbox_preds[0].dtype, device=bbox_preds[0].device
    )
    mlvl_priors = [priors.unsqueeze(0) for priors in mlvl_priors]

    mlvl_cls_scores = [cls_scores[i].detach() for i in range(num_levels)]
    mlvl_bbox_preds = [bbox_preds[i].detach() for i in range(num_levels)]
    if score_factors is None:
        with_score_factors = False
        mlvl_score_factor = [None for _ in range(num_levels)]
    else:
        with_score_factors = True
        mlvl_score_factor = [score_factors[i].detach() for i in range(num_levels)]
        mlvl_score_factors = []
    assert batch_img_metas is not None
    img_shape = batch_img_metas[0]["img_shape"]

    assert len(cls_scores) == len(bbox_preds) == len(mlvl_priors)
    batch_size = cls_scores[0].shape[0]
    cfg = self.test_cfg
    pre_topk = cfg.get("nms_pre", -1)

    mlvl_valid_bboxes = []
    mlvl_valid_scores = []
    mlvl_valid_priors = []

    for cls_score, bbox_pred, score_factors, priors, stride in zip(
        mlvl_cls_scores,
        mlvl_bbox_preds,
        mlvl_score_factor,
        mlvl_priors,
        self.prior_generator.strides,
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]
        assert stride[0] == stride[1]

        scores = cls_score.permute(0, 2, 3, 1).reshape(
            batch_size, -1, self.cls_out_channels
        )
        if self.use_sigmoid_cls:
            scores = scores.sigmoid()
            nms_pre_score = scores
        else:
            scores = scores.softmax(-1)
            nms_pre_score = scores
        if with_score_factors:
            score_factors = (
                score_factors.permute(0, 2, 3, 1).reshape(batch_size, -1).sigmoid()
            )
            score_factors = score_factors.unsqueeze(2)

        def _batched_integral(intergral, x):
            batch_size = x.size(0)
            x = F.softmax(x.reshape(batch_size, -1, intergral.reg_max + 1), dim=2)
            x = F.linear(x, intergral.project.type_as(x).unsqueeze(0)).reshape(
                batch_size, -1, 4
            )
            return x

        bbox_pred = (
            _batched_integral(self.integral, bbox_pred.permute(0, 2, 3, 1)) * stride[0]
        )
        if not is_dynamic_flag:
            priors = priors.data
        if pre_topk > 0:
            if with_score_factors:
                nms_pre_score = nms_pre_score * score_factors
            if backend == Backend.TENSORRT:
                priors = pad_with_value(priors, 1, pre_topk)
                bbox_pred = pad_with_value(bbox_pred, 1, pre_topk)
                scores = pad_with_value(scores, 1, pre_topk, 0.0)
                nms_pre_score = pad_with_value(nms_pre_score, 1, pre_topk, 0.0)
                if with_score_factors:
                    score_factors = pad_with_value(score_factors, 1, pre_topk, 0.0)

            # Get maximum scores for foreground classes.
            if self.use_sigmoid_cls:
                max_scores, _ = nms_pre_score.max(-1)
            else:
                max_scores, _ = nms_pre_score[..., :-1].max(-1)
            _, topk_inds = max_scores.topk(pre_topk)
            bbox_pred, scores, score_factors = gather_topk(
                bbox_pred,
                scores,
                score_factors,
                inds=topk_inds,
                batch_size=batch_size,
                is_batched=True,
            )
            priors = gather_topk(
                priors, inds=topk_inds, batch_size=batch_size, is_batched=False
            )

        mlvl_valid_bboxes.append(bbox_pred)
        mlvl_valid_scores.append(scores)
        priors = self.anchor_center(priors)
        mlvl_valid_priors.append(priors)
        if with_score_factors:
            mlvl_score_factors.append(score_factors)

    batch_mlvl_bboxes_pred = torch.cat(mlvl_valid_bboxes, dim=1)
    batch_scores = torch.cat(mlvl_valid_scores, dim=1)
    batch_priors = torch.cat(mlvl_valid_priors, dim=1)
    batch_bboxes = self.bbox_coder.decode(
        batch_priors, batch_mlvl_bboxes_pred, max_shape=img_shape
    )
    if with_score_factors:
        batch_score_factors = torch.cat(mlvl_score_factors, dim=1)

    if not self.use_sigmoid_cls:
        batch_scores = batch_scores[..., : self.num_classes]

    if with_score_factors:
        batch_scores = batch_scores * batch_score_factors
    if not with_nms:
        return batch_bboxes, batch_scores
    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        batch_bboxes,
        batch_scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.utils.gaussian_target.get_topk_from_heatmap"
)
def get_topk_from_heatmap__default(scores, k=20):
    """Get top k positions from heatmap.

    Replace view(batch, -1) with flatten
    """
    height, width = scores.size()[2:]
    topk_scores, topk_inds = torch.topk(scores.flatten(1), k)
    topk_clses = topk_inds // (height * width)
    topk_inds = topk_inds % (height * width)
    topk_ys = topk_inds // width
    topk_xs = (topk_inds % width).int().float()
    return topk_scores, topk_inds, topk_clses, topk_ys, topk_xs



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.openvino import is_available

__all__ = ["is_available"]

if is_available():
    from juxtapose.mmdeploy.backend.openvino.onnx2openvino import (
        from_onnx as _from_onnx,
    )
    from juxtapose.mmdeploy.backend.openvino.onnx2openvino import get_output_model_file
    from ..core import PIPELINE_MANAGER

    from_onnx = PIPELINE_MANAGER.register_pipeline()(_from_onnx)

    from .utils import get_input_info_from_cfg, get_mo_options_from_cfg

    __all__ += [
        "from_onnx",
        "get_output_model_file",
        "get_input_info_from_cfg",
        "get_mo_options_from_cfg",
    ]



# Copyright (c) OpenMMLab. All rights reserved.
import importlib
import logging
from abc import ABCMeta
from typing import Any, Callable, Optional, Sequence


class BaseBackendManager(metaclass=ABCMeta):
    """Abstract interface of backend manager."""

    @classmethod
    def build_wrapper(
        cls,
        backend_files: Sequence[str],
        device: str = "cpu",
        input_names: Optional[Sequence[str]] = None,
        output_names: Optional[Sequence[str]] = None,
        deploy_cfg: Optional[Any] = None,
        **kwargs,
    ):
        """Build the wrapper for the backend model.

        Args:
            backend_files (Sequence[str]): Backend files.
            device (str, optional): The device info. Defaults to 'cpu'.
            input_names (Optional[Sequence[str]], optional): input names.
                Defaults to None.
            output_names (Optional[Sequence[str]], optional): output names.
                Defaults to None.
            deploy_cfg (Optional[Any], optional): The deploy config. Defaults
                to None.
        """
        raise NotImplementedError(
            f"build_wrapper has not been implemented for `{cls.__name__}`"
        )

    @classmethod
    def is_available(cls, with_custom_ops: bool = False) -> bool:
        """Check whether backend is installed.

        Args:
            with_custom_ops (bool): check custom ops exists.
        Returns:
            bool: True if backend package is installed.
        """
        raise NotImplementedError(
            f'is_available has not been implemented for "{cls.__name__}"'
        )

    @classmethod
    def get_version(cls) -> str:
        """Get the version of the backend."""
        raise NotImplementedError(
            f'get_version has not been implemented for "{cls.__name__}"'
        )

    @classmethod
    def check_env(cls, log_callback: Callable = lambda _: _) -> str:
        """Check current environment.

        Returns:
            str: Info about the environment.
        """
        try:
            available = cls.is_available()
            if available:
                try:
                    backend_version = cls.get_version()
                except NotImplementedError:
                    backend_version = "Unknown"
            else:
                backend_version = "None"

            info = f"{cls.backend_name}:\t{backend_version}"
        except Exception:
            info = f"{cls.backend_name}:\tCheckFailed"

        log_callback(info)

        return info

    @classmethod
    def to_backend(
        cls,
        ir_files: Sequence[str],
        work_dir: str,
        deploy_cfg: Any,
        log_level: int = logging.INFO,
        device: str = "cpu",
        **kwargs,
    ) -> Sequence[str]:
        """Convert intermediate representation to given backend.

        Args:
            ir_files (Sequence[str]): The intermediate representation files.
            work_dir (str): The work directory, backend files and logs should
                be saved in this directory.
            deploy_cfg (Any): The deploy config.
            log_level (int, optional): The log level. Defaults to logging.INFO.
            device (str, optional): The device type. Defaults to 'cpu'.
        Returns:
            Sequence[str]: Backend files.
        """
        raise NotImplementedError(
            f"to_backend has not been implemented for `{cls.__name__}`"
        )


class BackendManagerRegistry:
    """backend manager registry."""

    def __init__(self):
        self._module_dict = {}

    def register(self, name: str, enum_name: Optional[str] = None):
        """register backend manager.

        Args:
            name (str): name of the backend
            enum_name (Optional[str], optional): enum name of the backend.
                if not given, the upper case of name would be used.
        """
        from juxtapose.mmdeploy.utils import get_root_logger

        logger = get_root_logger()

        if enum_name is None:
            enum_name = name.upper()

        def wrap_manager(cls):
            from juxtapose.mmdeploy.utils import Backend

            if not hasattr(Backend, enum_name):
                from aenum import extend_enum

                extend_enum(Backend, enum_name, name)
                logger.info(f"Registry new backend: {enum_name} = {name}.")

            if name in self._module_dict:
                logger.info(f"Backend manager of `{name}` has already been registered.")

            self._module_dict[name] = cls

            cls.backend_name = name

            return cls

        return wrap_manager

    def find(self, name: str) -> BaseBackendManager:
        """Find the backend manager with name.

        Args:
            name (str): backend name.
        Returns:
            BaseBackendManager: backend manager of the given backend.
        """
        # try import backend if backend is in `mmdeploy.backend`
        try:
            importlib.import_module("juxtapose.mmdeploy.backend." + name)
        except Exception:
            pass
        return self._module_dict.get(name, None)


BACKEND_MANAGERS = BackendManagerRegistry()


def get_backend_manager(name: str) -> BaseBackendManager:
    """Get backend manager.

    Args:
        name (str): name of the backend.
    Returns:
        BaseBackendManager: The backend manager of given name
    """
    from enum import Enum

    if isinstance(name, Enum):
        name = name.value
    return BACKEND_MANAGERS.find(name)



# https://github.com/open-mmlab/mmdeploy/blob/main/docs/en/04-supported-codebases/mmdet.md


from pathlib import Path
from juxtapose.mmdeploy.apis.utils import build_task_processor
from juxtapose.mmdeploy.utils import get_input_shape, load_config
from juxtapose.utils import get_user_config_dir
from juxtapose.utils.downloads import safe_download
from juxtapose.utils.core import Detections

base = "juxtapose"


class RTMDet:
    """RTMDet model (s, m, l) to detect only person (class 0)"""

    def __init__(self, type: str = "m", device: str = "cpu", conf_thres: float = 0.3):
        download_dir = Path("model")

        model_cfg = download_dir / f"rtmdet-{type}.py"
        onnx_file = download_dir / f"rtmdet-{type}.onnx"
        deploy_cfg = download_dir / f"detection_onnxruntime_static.py"

        if not model_cfg.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/rtmdet-{type}.py",
                file=f"rtmdet-{type}",
                dir=download_dir,
            )
        if not onnx_file.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/rtmdet-{type}.onnx",
                file=f"rtmdet-{type}",
                dir=download_dir,
            )
        if not deploy_cfg.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/detection_onnxruntime_static.py",
                file=f"detection_onnxruntime_static",
                dir=download_dir,
            )

        deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

        # build task and backend model
        self.task_processor = build_task_processor(model_cfg, deploy_cfg, device)
        self.model = self.task_processor.build_backend_model([onnx_file])

        # process input image
        self.input_shape = get_input_shape(deploy_cfg)

        self.conf_thres = conf_thres

    def __call__(self, im):
        """Return List of xyxy coordinates based on im frame (cv2.imread)
        im -> (h, w, c)
        return -> [[x, y, x, y], [x, y, x, y], ...] -> two persons detected
        """

        model_inputs, _ = self.task_processor.create_input(im, self.input_shape)
        result = self.model.test_step(model_inputs)

        pred_instances = result[0].pred_instances

        # filter confidence threshold
        pred_instances = pred_instances[pred_instances.scores > self.conf_thres]

        # get only class 0 (person)
        pred_instances = pred_instances[pred_instances.labels == 0].cpu().numpy()

        result = Detections(
            xyxy=pred_instances.bboxes,
            confidence=pred_instances.scores,
            labels=pred_instances.labels,
        )

        return result



# Copyright (c) OpenMMLab. All rights reserved.
from .object_detection import ObjectDetection
from .utils import (clip_bboxes, gather_topk, get_post_processing_params,
                    pad_with_value, pad_with_value_if_necessary)

__all__ = [
    'get_post_processing_params', 'clip_bboxes', 'pad_with_value',
    'pad_with_value_if_necessary', 'gather_topk', 'MMDetection',
    'ObjectDetection'
]



# Copyright (c) OpenMMLab. All rights reserved.
from . import bbox_head  # noqa: F401,F403
from . import cascade_roi_head  # noqa: F401,F403
from . import fcn_mask_head  # noqa: F401,F403
from . import single_level_roi_extractor  # noqa: F401,F403
from . import standard_roi_head  # noqa: F401,F403



# https://github.com/open-mmlab/mmdeploy/blob/main/docs/en/04-supported-codebases/mmpose.md

import numpy as np
from pathlib import Path

from juxtapose.utils import get_user_config_dir
from juxtapose.utils.downloads import safe_download
from juxtapose.mmdeploy.apis.utils import build_task_processor
from juxtapose.mmdeploy.utils import get_input_shape, load_config

from mmpose.structures import merge_data_samples


base = "juxtapose"


class RTMPose:
    """RTMPose model (s, m, l) to detect multi-person poses (class 0) based on bboxes"""

    def __init__(self, type: str = "m", device: str = "cpu") -> None:
        download_dir = Path("model")

        model_cfg = download_dir / f"rtmpose-{type}.py"
        onnx_file = download_dir / f"rtmpose-{type}.onnx"
        deploy_cfg = download_dir / f"pose-detection_simcc_onnxruntime_dynamic.py"

        if not model_cfg.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/rtmpose-{type}.py",
                file=f"rtmpose-{type}",
                dir=download_dir,
            )
        if not onnx_file.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/rtmpose-{type}.onnx",
                file=f"rtmpose-{type}",
                dir=download_dir,
            )
        if not deploy_cfg.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/pose-detection_simcc_onnxruntime_dynamic.py",
                file=f"detection_simcc_onnxruntime_dynamic",
                dir=download_dir,
            )

        # model_cfg = str(model_cfg)
        # onnx_file = str(onnx_file)
        # deploy_cfg = str(deploy_cfg)

        deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

        # build task and backend model
        self.task_processor = build_task_processor(model_cfg, deploy_cfg, device)
        self.model = self.task_processor.build_backend_model([onnx_file])

        # process input image
        self.input_shape = get_input_shape(deploy_cfg)

    def create_input(self, im, bboxes=None):
        model_inputs, _ = self.task_processor.create_input(
            im, self.input_shape, bboxes=bboxes
        )
        return model_inputs

    def __call__(self, im, bboxes=None):
        """Return List of 17 xy coordinates based on im frame (cv2.imread)
        im -> (h, w, c)
        bboxes -> [[x, y, x, y], [x, y, x, y], ...] -> two persons detected
        return -> np.ndarray([[x, y, ... 17 times], [x, y, ... 17 times], ...]) -> (2 or more, 17, 2)
        """

        if bboxes is not None:
            bboxes = [np.array([bbox]) for bbox in bboxes]

        model_inputs = self.create_input(im, bboxes)
        result = self.model.test_step(model_inputs)
        result = merge_data_samples(result)
        kpts = result.pred_instances.keypoints

        return kpts



# Copyright (c) OpenMMLab. All rights reserved.

import torch
from torch.types import Number

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.linspace")
def linspace__default(start: Number, end: Number, steps: int = None, **kwargs):
    """Rewrite `linspace` for onnxruntime."""
    steps = 100 if steps is None else steps
    dtype = kwargs.pop("dtype", torch.float32)
    dtype = dtype if dtype else torch.float32
    if steps == 1:
        output = torch.arange(start, end + 1, dtype=dtype, **kwargs)[:steps]
    else:
        output = torch.arange(
            start, end + 1, (end - start) / (steps - 1), dtype=dtype, **kwargs
        )[:steps]
    return output



_base_ = ['../../_base_/onnx_config.py']

onnx_config = dict(output_names=['dets', 'labels'], input_shape=None)
codebase_config = dict(
    type='mmdet',
    task='ObjectDetection',
    model_type='end2end',
    post_processing=dict(
        score_threshold=0.05,
        confidence_threshold=0.005,  # for YOLOv3
        iou_threshold=0.5,
        max_output_boxes_per_class=200,
        pre_top_k=5000,
        keep_top_k=100,
        background_label_id=-1,
    ))



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils.constants import Backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.prior_generators.MlvlPointGenerator"
    ".single_level_grid_priors",
    backend=Backend.TENSORRT.value,
)
def mlvl_point_generator__single_level_grid_priors__tensorrt(
    self, featmap_size, level_idx, dtype=torch.float32, device="cuda", with_stride=False
):
    """Rewrite `single_level_grid_priors` of `MlvlPointGenerator` as
    onnx2tensorrt raise the error of shape inference for YOLOX with some
    versions of TensorRT.

    Args:
        featmap_size (tuple[int]): Size of the feature maps, arrange as
            (h, w).
        level_idx (int): The index of corresponding feature map level.
        dtype (:obj:`dtype`): Dtype of priors. Default: torch.float32.
        device (str, optional): The device the tensor will be put on.
            Defaults to 'cuda'.
        with_stride (bool): Concatenate the stride to the last dimension
            of points.

    Return:
        Tensor: Points of single feature levels.
        The shape of tensor should be (N, 2) when with stride is
        ``False``, where N = width * height, width and height
        are the sizes of the corresponding feature level,
        and the last dimension 2 represent (coord_x, coord_y),
        otherwise the shape should be (N, 4),
        and the last dimension 4 represent
        (coord_x, coord_y, stride_w, stride_h).
    """
    feat_h, feat_w = featmap_size
    stride_w, stride_h = self.strides[level_idx]
    shift_x = (torch.arange(0, feat_w, device=device) + self.offset) * stride_w
    # keep featmap_size as Tensor instead of int, so that we
    # can convert to ONNX correctly
    shift_x = shift_x.to(dtype)

    shift_y = (torch.arange(0, feat_h, device=device) + self.offset) * stride_h
    # keep featmap_size as Tensor instead of int, so that we
    # can convert to ONNX correctly
    shift_y = shift_y.to(dtype)
    shift_xx, shift_yy = self._meshgrid(shift_x, shift_y)
    if not with_stride:
        shifts = torch.stack([shift_xx, shift_yy], dim=-1)
    else:
        # use `feat_w * feat_h` instead of `shift_xx.shape[0]` for TensorRT
        stride_w = shift_xx.new_full((feat_w * feat_h,), stride_w).to(dtype)
        stride_h = shift_xx.new_full((feat_w * feat_h,), stride_h).to(dtype)
        shifts = torch.stack([shift_xx, shift_yy, stride_w, stride_h], dim=-1)
    all_points = shifts.to(device)
    return all_points



# Copyright (c) OpenMMLab. All rights reserved.

from juxtapose.mmdeploy.backend.coreml import is_available

__all__ = ["is_available"]

if is_available():
    from juxtapose.mmdeploy.backend.coreml.torchscript2coreml import (
        from_torchscript as _from_torchscript,
    )
    from juxtapose.mmdeploy.backend.coreml.torchscript2coreml import get_model_suffix
    from ..core import PIPELINE_MANAGER

    from_torchscript = PIPELINE_MANAGER.register_pipeline()(_from_torchscript)
    __all__ += ["from_torchscript", "get_model_suffix"]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# Modified from:
# https://github.com/pytorch/pytorch/blob/9ade03959392e5a90b74261012de1d806cab2253/torch/onnx/symbolic_opset9.py
import warnings

import torch
import torch.onnx.symbolic_helper as sym_help
from torch.onnx.symbolic_helper import _unimplemented
from torch.onnx.symbolic_opset9 import unused

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.onnx.symbolic_opset9._generic_rnn", backend="ncnn"
)
def generic_rnn__ncnn(
    g,
    variant,
    input,
    initial_states,
    all_weights,
    has_biases,
    num_layers,
    dropout,
    train,
    bidirectional,
    batch_first=None,
    batch_sizes=None,
):
    """rewrite of _generic_rnn for ncnn.

    `g.op` will add some nodes for h0 and c0 in LSTM. which is not supported in
    ncnn. So we add a custom domain to avoid it.
    """
    warnings.warn(
        "Exporting a model to ONNX with a batch_size other than 1, "
        + "with a variable length with "
        + variant
        + " can cause an error "
        + "when running the ONNX model with a different batch size. "
        + "Make sure to save the model with a batch size of 1, "
        + "or define the initial states (h0/c0) as inputs of the model. "
    )

    onnxActivations = [
        "Relu",
        "Tanh",
        "Sigmoid",
        "Affine",
        "LeakyRelu",
        "ThresholdedRelu",
        "ScaledTanh",
        "HardSigmoid",
        "Elu",
        "Softsign",
        "Softplus",
    ]
    variantToOnnxActivationMap = dict(
        zip([act_fun.lower() for act_fun in onnxActivations], onnxActivations)
    )
    weights_per_layer = 4 if has_biases else 2
    # this means that projections are used inside LSTM,
    # so need to tell user that it's not supported
    if variant == "LSTM" and len(all_weights) != num_layers * weights_per_layer * (
        1 + bidirectional
    ):
        return _unimplemented("LSTM", "LSTMs with projections")
    assert len(all_weights) == num_layers * weights_per_layer * (1 + bidirectional)
    layer_weights = [
        all_weights[i : i + weights_per_layer]
        for i in range(0, len(all_weights), weights_per_layer)
    ]
    if batch_first:
        # batch, seq, feat -> seq, batch, feat
        input = g.op("Transpose", input, perm_i=[1, 0, 2])
    if dropout and train:
        return _unimplemented("RNN/GRU/LSTM", "dropout in training mode")

    if variant.startswith("RNN"):
        nonlinearity = variantToOnnxActivationMap[variant[4:].lower()]
        variant = "RNN"

    w_hh = all_weights[1]
    hidden_size = sym_help._get_tensor_dim_size(w_hh, 1)
    if hidden_size is None:
        return _unimplemented("RNN/GRU/LSTM", "unknown hidden size")

    unidirectional = not bidirectional

    prev_output = input

    h_outs = []
    if variant == "RNN" or variant == "GRU":
        h0 = initial_states
    elif variant == "LSTM":
        h0, c0 = initial_states
        c_outs = []

    sequence_lens = unused(g) if batch_sizes is None else batch_sizes

    if variant == "GRU":
        # pytorch is reset, input, hidden
        # onnx is    input, reset, hidden
        reform_permutation = [(1, 2), (0, 1), (2, 3)]
    elif variant == "LSTM":
        # pytorch is input, forget, cell, output.
        # onnx is    input, output, forget, cell.
        reform_permutation = [(0, 1), (3, 4), (1, 3)]

    def reform_weights(g, w, n, intervals):
        slices = [
            sym_help._slice_helper(g, w, axes=[0], starts=[x * n], ends=[y * n])
            for x, y in intervals
        ]
        return g.op("Concat", *slices, axis_i=0)

    def transform_weights_no_bias(layer_index):
        weights = layer_weights[layer_index]
        if variant == "RNN":
            weight_ih, weight_hh = weights
        elif variant == "GRU" or variant == "LSTM":
            weight_ih, weight_hh = [
                reform_weights(g, w, hidden_size, reform_permutation) for w in weights
            ]
        return tuple(
            sym_help._unsqueeze_helper(g, x, [0]) for x in (weight_ih, weight_hh)
        )

    def transform_weights(layer_index):
        weights = layer_weights[layer_index]
        if variant == "RNN":
            weight_ih, weight_hh, bias_ih, bias_hh = weights
        elif variant == "GRU" or variant == "LSTM":
            weight_ih, weight_hh, bias_ih, bias_hh = [
                reform_weights(g, w, hidden_size, reform_permutation) for w in weights
            ]
        bias_concat = g.op("Concat", bias_ih, bias_hh, axis_i=0)
        return tuple(
            sym_help._unsqueeze_helper(g, x, [0])
            for x in (weight_ih, weight_hh, bias_concat)
        )

    def retrieve_state(x, start, end):
        return (
            x
            if num_layers == 1
            else sym_help._slice_helper(g, x, axes=[0], starts=[start], ends=[end])
        )

    for i in range(num_layers):
        if unidirectional:
            if weights_per_layer == 4:
                weight_ih, weight_hh, bias_concat = transform_weights(i)
            else:
                weight_ih, weight_hh = transform_weights_no_bias(i)
                bias_concat = unused(g)

            state_indices = i, i + 1
        else:
            if weights_per_layer == 4:
                weight_ih_f, weight_hh_f, bias_f = transform_weights(2 * i)
                weight_ih_b, weight_hh_b, bias_b = transform_weights(2 * i + 1)
                bias_concat = g.op("Concat", bias_f, bias_b, axis_i=0)
            else:
                weight_ih_f, weight_hh_f = transform_weights_no_bias(2 * i)
                weight_ih_b, weight_hh_b = transform_weights_no_bias(2 * i + 1)
                bias_concat = unused(g)

            weight_ih = g.op("Concat", weight_ih_f, weight_ih_b, axis_i=0)
            weight_hh = g.op("Concat", weight_hh_f, weight_hh_b, axis_i=0)

            state_indices = 2 * i, 2 * i + 2

        inputs = [prev_output, weight_ih, weight_hh, bias_concat, sequence_lens]

        inputs.append(retrieve_state(h0, *state_indices))
        if variant == "LSTM":
            inputs.append(retrieve_state(c0, *state_indices))

        extra_kwargs = {} if unidirectional else {"direction_s": "bidirectional"}
        if variant == "RNN":
            if bidirectional:
                activation = [nonlinearity, nonlinearity]
            else:
                activation = [nonlinearity]

            prev_output, h_out = g.op(
                "RNN",
                *inputs,
                outputs=2,
                hidden_size_i=hidden_size,
                activations_s=activation,
                **extra_kwargs,
            )
        elif variant == "GRU":
            prev_output, h_out = g.op(
                "GRU",
                *inputs,
                outputs=2,
                hidden_size_i=hidden_size,
                linear_before_reset_i=1,
                **extra_kwargs,
            )
        elif variant == "LSTM":
            # g.op will add some node to h0 and c0,
            # which is not necessary for us
            prev_output, h_out, c_out = g.op(
                "ncnn::LSTM",
                *inputs,
                outputs=3,
                hidden_size_i=hidden_size,
                **extra_kwargs,
            )
        if bidirectional:
            # The ONNX RNN/GRU/LSTM produce an output of dimensions
            #   seq_len, num_directions, batch, hidden_size
            # We have to convert to match pytorch's expected
            #   seq_len, batch, num_directions * hidden_size
            # by first moving num_directions before hidden_size with
            # Transpose, and then combining it with hidden_size
            # with Reshape.
            prev_output = g.op("Transpose", prev_output, perm_i=[0, 2, 1, 3])
            prev_output = g.op(
                "Reshape",
                prev_output,
                g.op("Constant", value_t=torch.LongTensor([0, 0, -1])),
            )
        else:
            prev_output = sym_help._squeeze_helper(g, prev_output, [1])

        h_outs.append(h_out)
        if variant == "LSTM":
            c_outs.append(c_out)
    if batch_first:
        # seq, batch, num_directions * hidden_size -> batch, seq,
        # num_directions * hidden_size
        prev_output = g.op("Transpose", prev_output, perm_i=[1, 0, 2])
    h_outs = h_out if num_layers == 1 else g.op("Concat", *h_outs, axis_i=0)
    if variant == "RNN" or variant == "GRU":
        return prev_output, h_outs
    elif variant == "LSTM":
        c_outs = c_out if num_layers == 1 else g.op("Concat", *c_outs, axis_i=0)
        return prev_output, h_outs, c_outs



# Copyright (c) OpenMMLab. All rights reserved.
import re

import onnx
from packaging import version


def parse_extractor_io_string(io_str) -> tuple:
    """Parse IO string for extractor."""
    name, io_type = io_str.split(':')
    assert io_type in ['input', 'output']
    func_id = 0

    search_result = re.search(r'^(.+)\[([0-9]+)\]$', name)
    if search_result is not None:
        name = search_result.group(1)
        func_id = int(search_result.group(2))

    return name, func_id, io_type


def _dfs_search_reachable_nodes_fast(self, node_output_name, graph_input_nodes,
                                     reachable_nodes):
    """Using DFS to search reachable nodes."""
    outputs = {}
    for index, node in enumerate(self.graph.node):
        for name in node.output:
            if name not in outputs:
                outputs[name] = set()
            outputs[name].add(index)

    def impl(node_output_name, graph_input_nodes, reachable_nodes):
        if node_output_name in graph_input_nodes:
            return
        if node_output_name not in outputs:
            return
        for index in outputs[node_output_name]:
            node = self.graph.node[index]
            if node in reachable_nodes:
                continue
            reachable_nodes.append(node)
            for name in node.input:
                impl(name, graph_input_nodes, reachable_nodes)

    impl(node_output_name, graph_input_nodes, reachable_nodes)


def create_extractor(model: onnx.ModelProto) -> onnx.utils.Extractor:
    """Create Extractor for ONNX.

    Args:
        model (onnx.ModelProto): An input onnx model.

    Returns:
        onnx.utils.Extractor: Extractor for the onnx.
    """
    assert version.parse(onnx.__version__) >= version.parse('1.8.0')
    # patch extractor
    onnx.utils.Extractor._dfs_search_reachable_nodes = \
        _dfs_search_reachable_nodes_fast

    extractor = onnx.utils.Extractor(model)
    return extractor



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

__version__ = '1.2.0'
short_version = __version__


def parse_version_info(version_str: str) -> Tuple:
    """Parse version from a string.

    Args:
        version_str (str): A string represents a version info.

    Returns:
        tuple: A sequence of integer and string represents version.
    """
    version_info = []
    for x in version_str.split('.'):
        if x.isdigit():
            version_info.append(int(x))
        elif x.find('rc') != -1:
            patch_version = x.split('rc')
            version_info.append(int(patch_version[0]))
            version_info.append(f'rc{patch_version[1]}')
    return tuple(version_info)


version_info = parse_version_info(__version__)

__all__ = ['__version__', 'version_info', 'parse_version_info']
